#ifndef __ONIGURUMA_REGEX_H__
#define __ONIGURUMA_REGEX_H__

#include <ablib/KRsearch.h>
#include "Oniguruma_stdafx.h"

namespace Oniguruma{

class RegEx{
public:
	RegEx();
	~RegEx();
	RegEx(const TCHAR *reg);
	RegEx(const RegEx &obj);
	RegEx &operator=(const RegEx &obj);
	bool SetOpt(const TCHAR *set,bool val);
	bool Compile(const TCHAR *reg);
	bool Compile(const ablib::string &reg){return Compile(reg.c_str());}
	bool ReCompile(){return Compile(m_reg);}
	const TCHAR *GetSearchStr() const{return m_reg.c_str();}
	bool Match(const TCHAR *str);
	bool Match(const ablib::string &str){return Match(str.c_str());}

	const TCHAR *GetMatchStr() const{return m_match.str.c_str();}
	const TCHAR *GetMatchStr(int i);
	const TCHAR *GetMatchStr(const TCHAR *str);
	const TCHAR *GetMatchStr(const ablib::string &str){return GetMatchStr(str.c_str());}
	int GetMatchLen() const{return m_match.len;}
	int GetMatchLen(int i);
	int GetMatchPos(){return m_match.pos;}
	int GetMatchPos(int i);
	bool IsCompiled() const{return m_iscompiled;}
	int GetTokenNum(){return m_region->num_regs;}
	static const TCHAR *GetVer();
	const TCHAR *GetErrorMsg();
	
	// s̃}b`ׂgetsɕςD
	ablib::string Replace(const TCHAR *s,const TCHAR *gets);
	ablib::string Replace(const ablib::string &s,const TCHAR *gets){return Replace(s.c_str(),gets);}
	ablib::string Replace(const TCHAR *s,const ablib::string &gets){return Replace(s,gets.c_str());}
	ablib::string Replace(const ablib::string &s,const ablib::string &gets){return Replace(s.c_str(),gets.c_str());}

private:
	bool m_iscompiled;
	ablib::KRSearch m_noregsearch;
	ablib::string m_reg,m_target;
	OnigRegion *m_region;

	OnigErrorInfo m_err;
	int m_errcode;

	struct Option{
		Option() : CaseSense(true),word(false),zero_match(true),use_regular(true){}
		bool CaseSense;//啶ʂD
		bool word;//P̌D
		bool zero_match;//0̃}b`}b`Ƃ݂ȂD
		bool use_regular;//K\gD
		// SԂ̃IvVԂD
		OnigOptionType GetOnigurumaOption(){
			OnigOptionType rv = ONIG_OPTION_NONE;
			if(!CaseSense)ONIG_OPTION_ON(rv,ONIG_OPTION_IGNORECASE);
			if(!zero_match)ONIG_OPTION_ON(rv,ONIG_OPTION_FIND_NOT_EMPTY);
			return rv;
		}
	}m_opt;
	struct MatchData{
		ablib::string str;
		int pos;
		int len;
		bool ismatched;
	}m_match;

	regex_t *m_regex;


	bool IsEitango(TCHAR c){return (('a' <= c && c <= 'z') || ('A' <= c && c <= 'Z') || (c == '_'));}
#ifdef UNICODE
	int GetByteLen(const wchar_t *s){return 2*::lstrlenW(s);}
	int Byte2Len(int b){return b/2;}
	int Len2Byte(int l){return 2*l;}
#else
	int GetByteLen(const char *s){return ::lstrlenA(s);}
	int Byte2Len(int b){return b;}
	int Len2Byte(int l){return l;}
#endif

};

}//namespace Oniguruma

#endif //__ONIGURUMA_REGEX_H__

