#ifndef __ABLIB_KRSEARCH_H__
#define __ABLIB_KRSEARCH_H__

/** Karp-RabinASYɂ镶񌟍D
*/

namespace ablib{

namespace KR_detail{
	template<typename CH> int my_lstrlen(const CH* s);
	inline int my_lstrlen(const char *s){return ::lstrlenA(s);}
	inline int my_lstrlen(const wchar_t *s){return ::lstrlenW(s);}
}

template<class SEARCHCORE,typename CH> class BaseKR{
public:
	BaseKR(){}
	BaseKR(const ablib::basic_string<CH> &str){Compile(str);}
	~BaseKR(){}
	void Compile(const ablib::basic_string<CH> &str){m_search = str;m_powlen = pow2(m_search.length());}
	bool Match(const CH *str,int len){
		m_matchpos = SEARCHCORE::Search(str,len,m_search.c_str(),m_search.length(),m_powlen);
		return (m_matchpos >= 0);
	}
	bool Match(const CH *str){return Match(str,KR_detail::my_lstrlen(str));}
	bool Match(const ablib::basic_string<CH> &str){return Match(str.c_str(),str.length());}
	int GetMatchPos() const{return m_matchpos;}
	int GetMatchLen() const{return m_search.length();}
	const CH *GetMatchStr() const{return m_search.c_str();}
	const CH *GetSearchStr() const{return m_search.c_str();}
	static const TCHAR *strstr(const CH *str,const CH *pattern){return strstr(str,::lstrlen(str),pattern);}
	static const TCHAR *strstr(const CH *str,int strlen,const CH *pattern){
		int len = ::lstrlen(pattern);
		int rv = SEARCHCORE::Search(str,strlen,pattern,len,pow2(len));
		return (rv < 0 ? NULL : str + rv);
	}
	static TCHAR *strstr(CH *str,const CH *pattern){return strstr(str,::lstrlen(str),pattern);}
	static TCHAR *strstr(CH *str,int strlen,const CH *pattern){
		int len = ::lstrlen(pattern);
		int rv = SEARCHCORE::Search(str,strlen,pattern,len,pow2(len));
		return (rv < 0 ? NULL : str + rv);
	}


private:
	ablib::basic_string<CH> m_search;
	int m_matchpos;
	int m_powlen;// 2^{m_serach.length()}

	static inline int pow2(int n){return 1<<n;}
};

namespace KR_detail{
	template<typename CH> struct KRSearchCore{
		//d = 2^(searchlen)
		static int Search(const CH *target,int targetlen,const CH *search,int searchlen,int d){
			if(searchlen == 0)return 0;
			int targethash = Hash(target,searchlen);
			int searchhash = Hash(search,searchlen);
			int lasti = targetlen - searchlen;
			for(int i = 0 ; i <= lasti; ++i){
				if(targethash == searchhash && StrCheck(target + i,search,searchlen))return i;
				targethash = ReHash(target[i],target[i + searchlen],targethash,d);
			}
			return -1;
		}
		// âŌbD
		static inline int ReHash(CH a,CH b,int h,int d){return (h<<1)-a*d+b;}
	};
	template<typename CH> struct KRRSearchCore{
		//d = 2^(searchlen)
		static int Search(const CH *target,int targetlen,const CH *search,int searchlen,int d){
			if(searchlen == 0)return 0;
			int targethash = Hash(target + targetlen - searchlen,searchlen);
			int searchhash = Hash(search,searchlen);
			for(int i = targetlen - searchlen ; i > 0 ; --i){
				if(targethash == searchhash && StrCheck(target + i,search,searchlen))return i;
				targethash = ReHash(target[i + searchlen - 1],target[i - 1],targethash,d);
			}
			if(targethash == searchhash && StrCheck(target,search,searchlen))return 0;
			return -1;
		}
		// ŌâCbD
		static inline int ReHash(CH a,CH b,int h,int d){return ((h - a + b*d)>>1);}
	};
	template<typename CH> inline int Hash(const CH *s,int n){int h=0;for(int i=0;i<n;++i)h=(h<<1)+s[i];return h;}
	template<typename CH> inline bool StrCheck(const CH *target,const CH *search,int len){return ::memcmp(target,search,sizeof(CH)*len) == 0;}
}

typedef BaseKR<KR_detail::KRSearchCore<TCHAR>,TCHAR> KRSearch;
typedef BaseKR<KR_detail::KRRSearchCore<TCHAR>,TCHAR> KRRSearch;


}//namespace ablib

#endif //__ABLIB_KRSEARCH_H__

