#include "ablib.h"
#include "TextFile.h"
#include "IsKanji.h"
#include "SmallBuffer.h"

namespace ablib{

namespace{
	enum CodeSet{ ASCII, KANA, JIS};

	//JIS ZX 0208  Shift_JIS
	void jis2sjis(BYTE k,BYTE t,char *rv){
		if(k>=0x3f)	rv[0] = (char)(((k+1)>>1)+0xc0);
		else		rv[0] = (char)(((k+1)>>1)+0x80);
		if( k&1 )	rv[1] = (char)((t>>6) ? t+0x40 : t+0x3f);
		else		rv[1] = (char)(t+0x9e);
	}


	inline int GetMaskIndex(BYTE n){
		if( BYTE(n+2)  < 0xc2 ) return 1; // 00`10111111, fe, ff
		if( n          < 0xe0 ) return 2; // 110xxxxx
		if( n          < 0xf0 ) return 3; // 1110xxxx
		if( n          < 0xf8 ) return 4; // 11110xxx
		if( n          < 0xfc ) return 5; // 111110xx
		                        return 6; // 1111110x
	}
}

/** EUC̎D
* UNICODE`ĂꍇłShift_JISɂĂɂĂ̂ŁCJIS⏕R[h͎gȂD
*/
struct TextFile::rEUC : public TextFile::FileRead<BYTE>{
	CodeSet G[4];
	rEUC(FILE *fp,CodeSet g0,CodeSet g1,CodeSet g2,CodeSet g3) : FileRead<BYTE>(fp){
		G[0] = g0;G[1] = g1;G[2] = g2;G[3] = g3;
	}
	~rEUC(){}

	size_t tosjis(CodeSet cs,char *rv){
		BYTE c[2];
		switch(cs){
		case ASCII:
			rv[0] = GetChar();
			return 1;
		case JIS:
			c[0] = GetChar();c[1] = GetChar();
			jis2sjis((c[0] & 0x7F) - 0x20,(c[1] & 0x7F) - 0x20,rv);
			return 2;
		case KANA:
			rv[0] = GetChar();
			return 1;
		default:
			return 0;
		}
	}

	//VOVtgDEUC̓bLOVtgȂD
	static const BYTE SS2 = 0x8E;
	static const BYTE SS3 = 0x8F;

	size_t Read(TCHAR *buf,size_t size){return ReadImpl<false>(buf,size);}
	size_t ReadLine(TCHAR *buf,size_t size){return ReadImpl<true>(buf,size);}

#ifndef _UNICODE
	template<bool linebreak>size_t ReadImpl(TCHAR *cbuf,size_t size){
#else
	template<bool linebreak>size_t ReadImpl(TCHAR *buf,size_t size){
		char *cbuf = new char[size];
#endif
		size_t writesize = 0;
		while(!IsEof() && writesize < size - 2){
			BYTE c = GetOneChar();
			//sɒBD
			if(c == '\r' || c == '\n'){
				cbuf[writesize++] = '\n';
				if(GetChar() == '\r' && GetOneChar() == '\n')GetChar();
				if(linebreak)break;
			}
			switch(c){
			case SS2://VOVtg : G[2]
				GetChar();
				writesize += tosjis(G[2],cbuf + writesize);
				break;

			case SS3://VOVtg : G[3]
				GetChar();
				writesize += tosjis(G[3],cbuf + writesize);
				break;

			default:
				if(c >= 0x80){//GR = G[1]
					writesize += tosjis(G[1],cbuf + writesize);
				}else{//GL = G[0]
					writesize += tosjis(G[0],cbuf + writesize);
				}
			}
		}
		cbuf[writesize] = '\0';
#ifdef UNICODE
		writesize = ::MultiByteToWideChar(CP_ACP,MB_PRECOMPOSED,cbuf,writesize,buf,size);
		buf[writesize] = '\0';
		delete[] cbuf;
#endif
		return writesize;
	}
};

/** ISO-2022-JP̎DIso2022usASCIIvׂʂɂD
* ̂łD
*/
struct TextFile::rIsoJP : public TextFile::FileRead<BYTE>{
	CodeSet G0;
	rIsoJP(FILE *fp) : G0(ASCII),FileRead<BYTE>(fp){}
	~rIsoJP(){}

	const int ESC = 0x1B;

	void SwitchG(){
		BYTE b = GetChar();
		BYTE bb = GetChar();
		if(b == '('){
			if(bb == 'B' || bb == 'J')G0 = ASCII;
		}else if(b == '$'){
			if(bb == '@' || bb == 'B')G0 = JIS;
		}
	}

	size_t tosjis(CodeSet cs,char *rv){
		BYTE c[2];
		switch(cs){
		case ASCII:
			rv[0] = GetChar();
			return 1;
		case JIS:
			c[0] = GetChar();c[1] = GetChar();
			jis2sjis((c[0] & 0x7F) - 0x20,(c[1] & 0x7F) - 0x20,rv);
			return 2;
		case KANA:
			rv[0] = GetChar();
			return 1;
		default:
			return 0;
		}
	}

	size_t Read(TCHAR *buf,size_t size){return ReadImpl<false>(buf,size);}
	size_t ReadLine(TCHAR *buf,size_t size){return ReadImpl<true>(buf,size);}

#ifdef UNICODE
	template<bool linebreak>size_t ReadImpl(TCHAR  *buf,size_t size){
		char *cbuf = new char[size];
#else
	template<bool linebreak>size_t ReadImpl(TCHAR  *cbuf,size_t size){
#endif
		size_t writesize = 0;
		while(!IsEof() && writesize < size - 2){
			BYTE c = GetOneChar();
			if(c == ESC){
				GetChar();
				SwitchG();
			}else if(c == '\r' || c == '\n'){
				G0 = ASCII;//sł͕KASCIIɖ߂D
				cbuf[writesize++] = '\n';
				if(GetChar() == '\r' && GetOneChar() == '\n')GetChar();
				if(linebreak)break;
			}else writesize += tosjis(G0,cbuf + writesize);
		}
		cbuf[writesize] = '\0';
#ifdef UNICODE
		writesize = ::MultiByteToWideChar(CP_ACP,MB_PRECOMPOSED,cbuf,writesize,buf,size);
		buf[writesize] = '\0';
		delete[] cbuf;
#endif
		return writesize;
	}
};

/** Shift_JISǂݍ݁D
*/
struct TextFile::rSJIS : public TextFile::FileRead<char>{
	rSJIS(FILE *fp) : FileRead<char>(fp){}
	~rSJIS(){}

	size_t Read(TCHAR *buf,size_t size){return ReadImpl<false>(buf,size);}
	size_t ReadLine(TCHAR *buf,size_t size){return ReadImpl<true>(buf,size);}

#ifdef UNICODE
	template<bool linebreak>size_t ReadImpl(TCHAR  *buf,size_t size){
		char *cbuf = new char[size];
#else
	template<bool linebreak>size_t ReadImpl(TCHAR *cbuf,size_t size){
#endif
		size_t writesize = 0;
		while(!IsEof() && writesize < size - 2){
			char c = GetChar();
			if(c == '\r'){
				cbuf[writesize++] = '\n';
				if(GetOneChar() == '\n')GetChar();
				if(linebreak)break;
			}else if(c == '\n'){
				cbuf[writesize++] = '\n';
				if(linebreak)break;
			}else if(IsKanji(c)){
				cbuf[writesize++] = c;
				cbuf[writesize++] = GetChar();
			}else cbuf[writesize++] = c;
		}
		cbuf[writesize] = '\0';
#ifdef UNICODE
		writesize = ::MultiByteToWideChar(CP_ACP,MB_PRECOMPOSED,cbuf,writesize,buf,size);
		buf[writesize] = '\0';
		delete[] cbuf;
#endif
		return writesize;
	}
};

/** CWCuɔCĂ̂܂ܓǂݍރ[`DDSJIS + [CRLF] or [LF]pD
*/
struct TextFile::rDirect : public TextFile::READ{
	FILE *m_fp;
	rDirect(FILE *fp) : m_fp(fp){};
	~rDirect(){if(m_fp)::fclose(m_fp);}
	size_t ReadLine(TCHAR *buf,size_t size){
		if(::_fgetts(buf,(int)size,m_fp))return ::lstrlen(buf);
		else return 0;
	}
	size_t Read(TCHAR *buf,size_t size){
		size_t rv = ::fread(buf,sizeof(TCHAR),size - 2,m_fp);
		if(rv == 0){buf[0] = '\0';return 0;}
		// Sp1Byteڂs1Byteڂ肵1ByteǂݍށD
#ifndef UNICODE
		if((IsKanji(buf[rv - 1]) && !IsKanji2(buf,rv - 1)) || buf[rv - 1] == '\r'){
			rv += ::fread(buf + rv,sizeof(TCHAR),1,m_fp);
		}
#endif
		buf[rv] = '\0';
		return rv;
	}
};

/** UTF16ǂݍ݁DWindowsɊۓ邾EndianCBOM̈ɂ͋Ct悤D
*/
struct TextFile::rUTF16 : public FileRead<WORD>{
	rUTF16(FILE *fp,bool bigend) : FileRead<WORD>(fp){
		WORD w = GetOneChar();
		if(w == 0xFFFE || w == 0xFFFE0000){bigendian = true;GetChar();}
		else if(w == 0xFEFF){bigendian = false;GetChar();}
		else bigendian = bigend;
	}
	~rUTF16(){}

	WORD swap(WORD w) {return (w << 8) |(w >> 8);}
	
	size_t ReadLine(TCHAR *buf,size_t size){return ReadImpl<true>(buf,size);}
	size_t Read(TCHAR *buf,size_t size){return ReadImpl<false>(buf,size);}

#ifdef UNICODE
	template<bool linebreak> size_t ReadImpl(TCHAR *buf,size_t size){
#else 
	template<bool linebreak> size_t ReadImpl(TCHAR *cbuf,size_t csize){
		size_t size = (csize - 1)/2;
		wchar_t *buf = new wchar_t[size + 1];
#endif
		size_t writesize = 0;
		while(!IsEof() && writesize < size - 2){
			wchar_t c = bigendian ? swap(GetChar()) : GetChar();
			if(c == L'\r' || c == L'\n'){
				buf[writesize++] = L'\n';
				if(c == L'\r' && GetOneChar() == L'\n')GetChar();
				if(linebreak)break;
			}else buf[writesize++] = c;
		}
#ifndef UNICODE
		writesize = ::WideCharToMultiByte(CP_ACP,0,buf,(int)writesize,cbuf,(int)csize,NULL,NULL);
		cbuf[writesize] = '\0';
		delete[] buf;
#else
		buf[writesize] = '\0';
#endif
		return writesize;
	}
	bool bigendian;
};

/** UTF8D܂邿΂ƂƂ[[gۂD̂ň[UNICODEɂShift_JISցD
*/
struct TextFile::rUTF7_8 : public TextFile::FileRead<BYTE>{
	rUTF7_8(FILE *fp,int c) : FileRead<BYTE>(fp){
		switch(c){
		case UTF7	:cp = CP_UTF7;break;
		case UTF8	:
		case UTF8N	:cp = CP_UTF8;break;
		default:return;
		}
		if(cp == CP_UTF8){
			//BOMF
			if(GetChar() != 0xEF || GetChar() != 0xBB || GetChar() != 0xBF)Reset();
		}
	}
	~rUTF7_8(){}
	int cp;

	inline int GetNext(char c){
		if( BYTE(c+2) < 0xc2 )return 0; // 00`10111111, fe, ff
		if( c	      < 0xe0 )return 1; // 110xxxxx
		if( c         < 0xf0 )return 2; // 1110xxxx
		if( c         < 0xf8 )return 3; // 11110xxx
		if( c         < 0xfc )return 4; // 111110xx
		                      return 5; // 1111110x
	}

	size_t ReadLine(TCHAR *buf,size_t size){return ReadImpl<true>(buf,size);}
	size_t Read(TCHAR *buf,size_t size){return ReadImpl<false>(buf,size);}

#ifdef UNICODE
	template<bool linebreak> size_t ReadImpl(TCHAR *buf,size_t size){
		char *cbuf = new char[size];
#else 
	template<bool linebreak> size_t ReadImpl(TCHAR *cbuf,size_t size){
		wchar_t *buf = new wchar_t[size];
#endif
        size_t writesize = 0;
		while(!IsEof() && writesize < size - 2){
			BYTE c = GetChar();
			if(c == '\r' || c == '\n'){
				cbuf[writesize++] = '\n';
				if(c == '\r' && GetOneChar() == '\n')GetChar();
				if(linebreak)break;
			}else{
				cbuf[writesize++] = c;
				if(c & 0x80){
					int n = GetNext(c);
					for(int i = 0 ; i < n ; ++i)cbuf[writesize++] = GetChar();
				}
			}
		}
		writesize = ::MultiByteToWideChar(cp,0,cbuf,(int)writesize,buf,(int)size);
#ifdef UNICODE
		buf[writesize] = '\0';
		delete[] cbuf;
#else
		writesize = ::WideCharToMultiByte(CP_ACP,0,buf,(int)writesize,cbuf,(int)size,NULL,NULL);
		cbuf[writesize] = '\0';
		delete[] buf;
#endif 
		return writesize;
	}
};

namespace{
	//SJIS ==> JIS X 0208
	static void sjis2jis(BYTE s1,BYTE s2,BYTE *k ){
		if( ((s1==0xfa || s1==0xfb) && s2>=0x40) || (s1==0xfc && (s2&0xf0)==0x40) ){
			// IBMÕ}bsO
			static const WORD IBM_sjis2kuten[] = {
			/*fa40*/0x5c51,0x5c52,0x5c53,0x5c54,0x5c55,0x5c56,0x5c57,0x5c58,0x5c59,0x5c5a,0x0d15,0x0d16,0x0d17,0x0d18,0x0d19,0x0d1a,
			/*fa50*/0x0d1b,0x0d1c,0x0d1d,0x0d1e,0x022c,0x5c5c,0x5c5d,0x5c5e,0x0d4a,0x0d42,0x0d44,0x0248,0x5901,0x5902,0x5903,0x5904,
			/*fa60*/0x5905,0x5906,0x5907,0x5908,0x5909,0x590a,0x590b,0x590c,0x590d,0x590e,0x590f,0x5910,0x5911,0x5912,0x5913,0x5914,
			/*fa70*/0x5915,0x5916,0x5917,0x5918,0x5919,0x591a,0x591b,0x591c,0x591d,0x591e,0x591f,0x5920,0x5921,0x5922,0x5923,/**/0x0109,
			/*fa80*/0x5924,0x5925,0x5926,0x5927,0x5928,0x5929,0x592a,0x592b,0x592c,0x592d,0x592e,0x592f,0x5930,0x5931,0x5932,0x5933,
			/*fa90*/0x5934,0x5935,0x5936,0x5937,0x5938,0x5939,0x593a,0x593b,0x593c,0x593d,0x593e,0x593f,0x5940,0x5941,0x5942,0x5943,
			/*faa0*/0x5944,0x5945,0x5946,0x5947,0x5948,0x5949,0x594a,0x594b,0x594c,0x594d,0x594e,0x594f,0x5950,0x5951,0x5952,0x5953,
			/*fab0*/0x5954,0x5955,0x5956,0x5957,0x5958,0x5959,0x595a,0x595b,0x595c,0x595d,0x595e,0x5a01,0x5a02,0x5a03,0x5a04,0x5a05,
			/*fac0*/0x5a06,0x5a07,0x5a08,0x5a09,0x5a0a,0x5a0b,0x5a0c,0x5a0d,0x5a0e,0x5a0f,0x5a10,0x5a11,0x5a12,0x5a13,0x5a14,0x5a15,
			/*fad0*/0x5a16,0x5a17,0x5a18,0x5a19,0x5a1a,0x5a1b,0x5a1c,0x5a1d,0x5a1e,0x5a1f,0x5a20,0x5a21,0x5a22,0x5a23,0x5a24,0x5a25,
			/*fae0*/0x5a26,0x5a27,0x5a28,0x5a29,0x5a2a,0x5a2b,0x5a2c,0x5a2d,0x5a2e,0x5a2f,0x5a30,0x5a31,0x5a32,0x5a33,0x5a34,0x5a35,
			/*faf0*/0x5a36,0x5a37,0x5a38,0x5a39,0x5a3a,0x5a3b,0x5a3c,0x5a3d,0x5a3e,0x5a3f,0x5a40,0x5a41,0x5a42,/**/0x0109,0x0109,0x0109,
			/*fb40*/0x5a43,0x5a44,0x5a45,0x5a46,0x5a47,0x5a48,0x5a49,0x5a4a,0x5a4b,0x5a4c,0x5a4d,0x5a4e,0x5a4f,0x5a50,0x5a51,0x5a52,
			/*fb50*/0x5a53,0x5a54,0x5a55,0x5a56,0x5a57,0x5a58,0x5a59,0x5a5a,0x5a5b,0x5a5c,0x5a5d,0x5a5e,0x5b01,0x5b02,0x5b03,0x5b04,
			/*fb60*/0x5b05,0x5b06,0x5b07,0x5b08,0x5b09,0x5b0a,0x5b0b,0x5b0c,0x5b0d,0x5b0e,0x5b0f,0x5b10,0x5b11,0x5b12,0x5b13,0x5b14,
			/*fb70*/0x5b15,0x5b16,0x5b17,0x5b18,0x5b19,0x5b1a,0x5b1b,0x5b1c,0x5b1d,0x5b1e,0x5b1f,0x5b20,0x5b21,0x5b22,0x5b23,/**/0x0109,
			/*fb80*/0x5b24,0x5b25,0x5b26,0x5b27,0x5b28,0x5b29,0x5b2a,0x5b2b,0x5b2c,0x5b2d,0x5b2e,0x5b2f,0x5b30,0x5b31,0x5b32,0x5b33,
			/*fb90*/0x5b34,0x5b35,0x5b36,0x5b37,0x5b38,0x5b39,0x5b3a,0x5b3b,0x5b3c,0x5b3d,0x5b3e,0x5b3f,0x5b40,0x5b41,0x5b42,0x5b43,
			/*fba0*/0x5b44,0x5b45,0x5b46,0x5b47,0x5b48,0x5b49,0x5b4a,0x5b4b,0x5b4c,0x5b4d,0x5b4e,0x5b4f,0x5b50,0x5b51,0x5b52,0x5b53,
			/*fbb0*/0x5b54,0x5b55,0x5b56,0x5b57,0x5b58,0x5b59,0x5b5a,0x5b5b,0x5b5c,0x5b5d,0x5b5e,0x5c01,0x5c02,0x5c03,0x5c04,0x5c05,
			/*fbc0*/0x5c06,0x5c07,0x5c08,0x5c09,0x5c0a,0x5c0b,0x5c0c,0x5c0d,0x5c0e,0x5c0f,0x5c10,0x5c11,0x5c12,0x5c13,0x5c14,0x5c15,
			/*fbd0*/0x5c16,0x5c17,0x5c18,0x5c19,0x5c1a,0x5c1b,0x5c1c,0x5c1d,0x5c1e,0x5c1f,0x5c20,0x5c21,0x5c22,0x5c23,0x5c24,0x5c25,
			/*fbe0*/0x5c26,0x5c27,0x5c28,0x5c29,0x5c2a,0x5c2b,0x5c2c,0x5c2d,0x5c2e,0x5c2f,0x5c30,0x5c31,0x5c32,0x5c33,0x5c34,0x5c35,
			/*fbf0*/0x5c36,0x5c37,0x5c38,0x5c39,0x5c3a,0x5c3b,0x5c3c,0x5c3d,0x5c3e,0x5c3f,0x5c40,0x5c41,0x5c42,/**/0x0109,0x0109,0x0109,
			/*fc40*/0x5c43,0x5c44,0x5c45,0x5c46,0x5c47,0x5c48,0x5c49,0x5c4a,0x5c4b,0x5c4c,0x5c4d,0x5c4e,/**/0x0109,0x0109,0x0109,0x0109,
			};
			k[0] = IBM_sjis2kuten[ (s1-0xfa)*12*16 + (s2-0x40) ]>>8;
			k[1] = IBM_sjis2kuten[ (s1-0xfa)*12*16 + (s2-0x40) ]&0xff;
		}else{
			// ̑
			if( s2>=0x9f )
			{
				if( s1>=0xe0 ) k[0] = ((s1-0xc0)<<1);
				else           k[0] = ((s1-0x80)<<1);
				k[1] = s2-0x9e;
			}
			else
			{
				if( s1>=0xe0 ) k[0] = ((s1-0xc0)<<1)-1;
				else           k[0] = ((s1-0x80)<<1)-1;

				if( s2 & 0x80 )	k[1] = s2-0x40;
				else			k[1] = s2-0x3f;
			}
		}

		k[0] += 0x20;
		k[1] += 0x20;
	}
}


/** EUC݃[`DShift_JIS炾ƎvĂ邩VtgȂ̂ŊȒPD
*/
struct TextFile::wEUC : public FileWrite<BYTE>{
	wEUC(FILE *fp,int l) : FileWrite<BYTE>(fp){
		switch(l){
		case CR:   ::lstrcpyA(lb,"\r");		break;
		case LF:   ::lstrcpyA(lb,"\n");		break;
		case CRLF: ::lstrcpyA(lb,"\r\n");	break;
		default:   ASSERT(false);			break;//Ȃ͂D
		}
	}
	~wEUC(){}

	//sێD
	char lb[3];

#ifdef UNICODE
	size_t Write(const TCHAR *buf,size_t size){
		int bufsize = ::WideCharToMultiByte(CP_ACP,0,buf,size,NULL,0,NULL,NULL) + 1;
		char *cbuf = new char[bufsize];
		size = ::WideCharToMultiByte(CP_ACP,0,buf,size,cbuf,bufsize,NULL,NULL);
#else
	size_t Write(const TCHAR *cbuf,size_t size){
#endif
		const char *end = cbuf + size;
		size_t writesize = 0;
		BYTE tmp[3];
		while(cbuf < end){
			if(*cbuf == '\r' || *cbuf == '\n'){
				for(char *p = lb ; *p ; ++p){WriteChar(*p);++writesize;}
				if(*cbuf == '\r' && *(cbuf + 1) == '\n')++cbuf;
				++cbuf;
			}else if(IsKanji(*cbuf)){
				sjis2jis(*cbuf,*(cbuf + 1),tmp);
				WriteChar(tmp[0] | 0x80);WriteChar(tmp[1] | 0x80);
				writesize += 2;cbuf += 2;
			}else{
				WriteChar(*(cbuf++));
				writesize += 1;
			}
		}
#ifdef UNICODE
		delete[] (end - size);
#endif
		return writesize;
	}
};

/** ISO-2022-JP݃[`DĔpJi͖ijD
*/
struct TextFile::wIsoJP : public TextFile::FileWrite<BYTE>{
	CodeSet G0;
	wIsoJP(FILE *fp,int l) : G0(ASCII),FileWrite<BYTE>(fp){
		switch(l){
		case CR:   ::lstrcpyA(lb,"\r");		break;
		case LF:   ::lstrcpyA(lb,"\n");		break;
		case CRLF: ::lstrcpyA(lb,"\r\n");	break;
		default:   ASSERT(false);			break;//Ȃ͂D
		}
	}
	~wIsoJP(){}

	//sێD
	char lb[3];

	const int ESC = 0x1B;

	size_t WriteESC(CodeSet c){
		WriteChar(ESC);
		G0 = c;
		switch(c){
		case ASCII:
			WriteChar('(');WriteChar('B');
			return 3;
		case JIS:
			WriteChar('$');WriteChar('B');
			return 3;
		default:
			return 1;
		}

	}


#ifdef UNICODE
	size_t Write(const TCHAR *buf,size_t size){
		int bufsize = ::WideCharToMultiByte(CP_ACP,0,buf,size,NULL,0,NULL,NULL) + 1;
		char *cbuf = new char[bufsize];
		size = ::WideCharToMultiByte(CP_ACP,0,buf,size,cbuf,bufsize,NULL,NULL);
#else
	size_t Write(const TCHAR *cbuf,size_t size){
#endif
		const char *end = cbuf + size;
		size_t writesize = 0;
		BYTE tmp[3];
		while(cbuf < end){
			if(*cbuf == '\r' || *cbuf == '\n'){
				if(G0 != ASCII)writesize += WriteESC(ASCII);
				for(char *p = lb ; *p ; ++p){WriteChar(*p);++writesize;}
				if(*cbuf == '\r' && *(cbuf + 1) == '\n')++cbuf;
				++cbuf;
			}else if(IsKanji(*cbuf)){
				if(G0 != JIS)writesize += WriteESC(JIS);
				sjis2jis(*cbuf,*(cbuf + 1),tmp);
				WriteChar(tmp[0]);WriteChar(tmp[1]);
				writesize += 2;cbuf += 2;
			}else{
				if(G0 != ASCII)writesize += WriteESC(ASCII);
				WriteChar(*(cbuf++));
				writesize += 1;
			}
		}
#ifdef UNICODE
		delete[] (end - size);
#endif
		return writesize;
	}
};

/** SJIS݃[`DȒP[D
*/
struct TextFile::wSJIS : public TextFile::FileWrite<BYTE>{
	wSJIS(FILE *fp,int l) : FileWrite<BYTE>(fp){
		switch(l){
		case CR:   ::lstrcpyA(lb,"\r");  break;
		case LF:   ::lstrcpyA(lb,"\n");  break;
		case CRLF: ::lstrcpyA(lb,"\r\n");break;
		default:   ::lstrcpyA(lb,"");    break;//Ȃ͂D
		}
	}
	~wSJIS(){}

	//sێD
	char lb[3];

#ifdef UNICODE
	size_t Write(const TCHAR *buf,size_t size){
		int bufsize = ::WideCharToMultiByte(CP_ACP,0,buf,size,NULL,0,NULL,NULL) + 1;
		char *cbuf = new char[bufsize];
		size = ::WideCharToMultiByte(CP_ACP,0,buf,size,cbuf,bufsize,NULL,NULL);
		cbuf[size] = '\0';
#else
	size_t Write(const TCHAR *cbuf,size_t size){
#endif
		const char *end = cbuf + size;
		size_t writesize = 0;
		while(cbuf < end){
			if(*cbuf == '\r' || *cbuf == '\n'){
				for(char *p = lb ; *p ; ++p){WriteChar(*p);++writesize;}
				if(*cbuf == '\r' && *(cbuf + 1) == '\n')++cbuf;
				++cbuf;
			}else if(IsKanji(*cbuf)){
				WriteChar(*(cbuf++));WriteChar(*(cbuf++));
				writesize += 2;
			}else{
				WriteChar(*(cbuf++));
				writesize += 1;
			}
		}
#ifdef UNICODE
		delete[] (end - size);
#endif
		return writesize;
	}

};

/** CWCuɂ܂Ă̂܂܏ރ[`D
*/
struct TextFile::wDirect : public WRITE{
	FILE *m_fp;
	wDirect(FILE *fp) : m_fp(fp){};
	~wDirect(){if(m_fp)::fclose(m_fp);}

	size_t Write(const TCHAR *buf,size_t size){return ::fwrite(buf,sizeof(TCHAR),size,m_fp);}
};

struct TextFile::wUTF16 : public TextFile::FileWrite<WORD>{
	wUTF16(FILE *fp,int l,bool bigend,bool bom) : FileWrite<WORD>(fp){
		switch(l){
		case CR		: ::lstrcpyW(lb,L"\r");		break;
		case LF		: ::lstrcpyW(lb,L"\n");		break;
		case CRLF	: ::lstrcpyW(lb,L"\r\n");	break;
		default		: ::lstrcpyW(lb,L"");		break;//蓾Ȃ
		}
		bigendian = bigend;
		if(bom)WriteChar(bigendian ? 0xFFFE : 0xFEFF);
	}
	wchar_t lb[3];
	bool bigendian;

	WORD  swap(WORD w) {return (w << 8) |(w >> 8);}

#ifdef UNICODE
	size_t Write(const TCHAR *buf,size_t size){
#else 
	size_t Write(const TCHAR *cbuf,size_t size){
		wchar_t *buf = new wchar_t[size];
		size = ::MultiByteToWideChar(CP_ACP,0,cbuf,(int)size,buf,(int)size);
#endif
		const wchar_t *start = buf;
		const wchar_t *end = buf + size;
		size_t writesize = 0;
		while(buf < end){
			if(*buf == '\r' || *buf == '\n'){
				for(const wchar_t *p = lb ; *p ; ++p){WriteChar(bigendian ? swap(*p) : *p);++writesize;}
				if(*buf == '\r' && *(buf + 1) == '\n')++buf;
				++buf;
			}else{
				WriteChar(bigendian ? swap(*(buf++)) : *(buf++));
				++writesize;
			}
		}
#ifndef UNICODE
		delete[] start;
#endif
		return writesize;
	}
};

struct TextFile::wUTF7_8 : public TextFile::FileWrite<BYTE>{
	wUTF7_8(FILE *fp,int l,int c) : FileWrite<BYTE>(fp){
		switch(l){
		case CR		: ::lstrcpyA(lb,"\r");	break;
		case LF		: ::lstrcpyA(lb,"\n");	break;
		case CRLF	: ::lstrcpyA(lb,"\r\n");break;
		}
		switch(c){
		case UTF8	:
		case UTF8N	:cp = CP_UTF8;break;
		case UTF7	:cp = CP_UTF7;break;
		}
		if(c == UTF8){
			WriteChar(0xEF);WriteChar(0xBB);WriteChar(0xBF);
		}
	}

	char lb[3];
	int cp;
	ablib::SmallBuffer<char> cbuffer;
#ifndef UNICODE
	ablib::SmallBuffer<wchar_t> wbuffer;
#endif

	size_t Write(const TCHAR *b,size_t s){
#ifdef UNICODE
		cbuffer.Alloc(s * 6 + 2);
		size_t size = ::WideCharToMultiByte(cp, 0, b, s, cbuffer.buf, cbuffer.size, NULL, NULL);
		OUTPUTLASTERROR();
#else
		wbuffer.Alloc(s);
		size_t size = ::MultiByteToWideChar(CP_ACP, 0, b, (int)s, wbuffer.buf, (int)wbuffer.size);
		cbuffer.Alloc(size * 6);
		size = ::WideCharToMultiByte(cp, 0, wbuffer.buf, (int)size, cbuffer.buf, (int)cbuffer.size, NULL, NULL);
		DWORD d = ::GetLastError();
#endif
		size_t writesize = 0;
		char *end = cbuffer.buf + size;
		char *buf = cbuffer.buf;
		while(buf < end){
			if(*buf == '\r' || *buf == '\n'){
				for(int i = 0 ; i < ::lstrlenA(lb) ; ++i){WriteChar(lb[i]);++writesize;}
				if (*buf == '\r' && *(buf + 1) == '\n')++buf;
				++buf;
			}else{
				WriteChar(*(buf++));++writesize;
			}
		}
		return writesize;
	}

};

/** t@CI[vDsR[h͓ǂݍݎɂ͖wǖD
*/
bool TextFile::Open(const TCHAR *file,filemode mode,int cs,int lb){
	if(file == NULL)return false;
	if(*file == '\0')return false;
	if(m_Read || m_Write)return false;
	if(mode == append)return false;//
	FILE *fp;
	if(mode == read){
		fp = ::_tfopen(file,_T("rb"));
		if(fp == NULL)return false;
		BYTE b[16384];
		size_t size = ::fread(b,sizeof(BYTE),16384,fp);
		AutoDetection(b,size,cs,lb);
		::fseek(fp,0,SEEK_SET);
		if(cs != AutoDetect){
			m_cs = cs;
		}
#ifndef UNICODE
		if(m_cs == SJIS && m_lb != CR){
			::fclose(fp);
			fp = ::_tfopen(file,_T("r"));
			if(fp == NULL)return false;
			m_Read = new rDirect(fp);
			return true;
		}
#endif
		switch(m_cs){
		case UTF7	:
		case UTF8	:
		case UTF8N	:m_Read = new rUTF7_8(fp,m_cs);				break;
		case UTF16b	:
		case UTF16BE:m_Read = new rUTF16(fp,true);				break;
		case UTF16l	:
		case UTF16LE:m_Read = new rUTF16(fp,false);				break;
//		case UTF32b	:
//		case UTF32BE:m_Read = new rUTF32(fp,true);				break;
//		case UTF32l	:
//		case UTF32LE:m_Read = new rUTF32(fp,false);				break;
		case EucJP  :m_Read = new rEUC(fp,ASCII,JIS,KANA,JIS);  break;
		case IsoJP  :m_Read = new rIsoJP(fp);                   break;
		case SJIS   :
		default     :m_Read = new rSJIS(fp);                    break;
		}
		return true;
	}else if(mode == write){
		if(cs == AutoDetect){
#ifdef UNICODE
			cs = UTF16l;
#else
			cs = SJIS;
#endif
		}
		m_cs = cs;m_lb = lb;
#ifndef UNICODE
		if(m_cs == SJIS && m_lb == CRLF){
			fp = ::_tfopen(file,_T("w"));
			if(fp == NULL)return false;
			m_Write = new wDirect(fp);
			return true;
		}
#endif
		fp = ::_tfopen(file,_T("wb"));
		if(fp == NULL)return false;
		switch(m_cs){
		case UTF7	:
		case UTF8	:
		case UTF8N	:m_Write = new wUTF7_8(fp,lb,m_cs);			break;
		case UTF16b	:m_Write = new wUTF16(fp,lb,true,true);		break;
		case UTF16BE:m_Write = new wUTF16(fp,lb,true,false);	break;
		case UTF16l	:m_Write = new wUTF16(fp,lb,false,true);	break;
		case UTF16LE:m_Write = new wUTF16(fp,lb,false,false);	break;
		case EucJP  :m_Write = new wEUC(fp,lb);					break;
		case IsoJP  :m_Write = new wIsoJP(fp,lb);				break;
		case SJIS   :
		default     :m_Write = new wSJIS(fp,lb);				break;
		}
		return true;
	}
	return false;
}

/** sR[hyѕR[h𔻕ʂDK.INABAɂD
*/
void TextFile::AutoDetection(BYTE *ptr,size_t size,int defcs,int deflb){
//-- ܂Ȁo񐔂̓v
	m_cs = defcs;
	m_lb = deflb;
	int  freq[256];
	bool bit8 = false;
	::ZeroMemory(freq,sizeof(int) * 256);
	for( ULONG i=0; i<size; ++i ){
		if( ptr[i] >= 0x80 )
			bit8 = true;
		++freq[ ptr[i] ];
	}

//-- sR[h (UTF16/32/7̂Ƃ肠BUTF5ɎĂ͔sc)

	if( freq['\r'] > freq['\n']*2 )		m_lb = CR;
	else if( freq['\n'] > freq['\r']*2 )m_lb = LF;
	else                                m_lb = CRLF;

//-- ftHgR[h

	int defc,defl;
	defc = Western;
	defl = m_lb;
	switch(::GetACP()){
	case 932: defc = SJIS;    break;
	case 936: defc = GBK;     break;
	case 949: defc = UHC;     break;
	case 950: defc = Big5;    break;
	}

//-- ꍇ͂ŏI

	if( size <= 4 ){
		if(defcs == AutoDetect){m_cs = defc;m_lb = defl;}
		else m_lb = deflb;
		return;
	}

//-- w肪ꍇ͂ŏI

	ULONG bom4 = (ptr[0]<<24) + (ptr[1]<<16) + (ptr[2]<<8) + (ptr[3]);
	ULONG bom2 = (ptr[0]<<8)  + (ptr[1]);

	if( m_cs==UTF8 || m_cs==UTF8N )m_cs = (bom4>>8==0xefbbbf ? UTF8 : UTF8N);
	else if( m_cs==UTF32b || m_cs==UTF32BE )m_cs = (bom4==0x0000feff ? UTF32b : UTF32BE);
	else if( m_cs==UTF32l || m_cs==UTF32LE )m_cs = (bom4==0xfffe0000 ? UTF32l : UTF32LE);
	else if( m_cs==UTF16b || m_cs==UTF16BE )m_cs = (bom2==0xfeff ? UTF16b : UTF16BE);
	else if( m_cs==UTF16l || m_cs==UTF16LE )m_cs = (bom2==0xfffe ? UTF16l : UTF16LE);

	if( m_cs != AutoDetect )return;

//-- BOM`FbNE7bit`FbN

	bool Jp = ::IsValidCodePage(932)!=FALSE;

	if( (bom4>>8) == 0xefbbbf )   m_cs = UTF8;
	else if( bom4 == 0x0000feff ) m_cs = UTF32b;
	else if( bom4 == 0xfffe0000 ) m_cs = UTF32l;
	else if( bom2 == 0xfeff )     m_cs = UTF16b;
	else if( bom2 == 0xfffe )     m_cs = UTF16l;
	else if( bom4 == 0x1b242943 && ::IsValidCodePage(949) ) m_cs = IsoKR;
	else if( bom4 == 0x1b242941 && ::IsValidCodePage(936) ) m_cs = IsoCN;
	else if( Jp && !bit8 && freq[0x1b]>0 )                  m_cs = IsoJP;

	if( m_cs != AutoDetect )return;

//-- UTF-5 `FbN

	ULONG u5sum = 0;
	for(BYTE c = '0'; c <= '9' ; ++c) u5sum += freq[c];
	for(BYTE c = 'A'; c <= 'V' ; ++c) u5sum += freq[c];
	if(size == u5sum){
		m_cs = UTF5;
		return;
	}

//-- b UTF-8 / {EUC `FbN

	m_cs = defc;

	// sR[hLFAx̑傫AłȂ
	//  ANSI-CP ƌȂĂ܂B
	if( bit8 && (size>4096 || m_lb==1
	 || freq[0xfd]>0 || freq[0xfe]>0 || freq[0xff]>0 || freq[0x80]>0) ){
		if( Jp ){
			// EUCƂĂl`FbN
			bool be=true;
			for( int k=0x90; k<=0xa0; ++k )if( freq[k]>0 ){be=false;break;}
			for( int k=0x7f; k<=0x8d; ++k )if( freq[k]>0 ){be=false;break;}
			if( be ){
				m_cs = EucJP;
			}
		}
		{
			// UTF8Ƃēǂ߂邩ǂ`FbN
			bool b8=true;
			int mi=1;
			for( ULONG i=0; i<size && b8; ++i )
				if( --mi ){
					if( ptr[i]<0x80 || ptr[i]>=0xc0 )
						b8 = false;
				}else{
					mi = 1;
					if( ptr[i] > 0x7f ){
						mi = GetMaskIndex( ptr[i] );
						if( mi == 1 )//ptr[i] >= 0xfe )
							b8 = false;
					}
				}
				if( b8 ){
					m_cs = UTF8N;
					return;
				}
		}
	}
	return;
}

}//namespace ablib


