#ifndef __ABLIB_LIST_H__
#define __ABLIB_LIST_H__

#include "ablib.h"
#include "alloc.h"

namespace ablib{

/** std::list̋UD
*/

template<typename T , typename Alloc = allocator<T> >class list{
public:
	typedef size_t size_type;
	list() : m_size(0){
		m_tail = m_nodealloc.allocate(1);
		m_nodealloc.construct(m_tail,_Node());
		m_head = m_tail;
	};
	~list(){
		_Node *p = m_head;
		while(p){
			_Node *q = p->next;
			m_nodealloc.destroy(p);
			m_nodealloc.deallocate(p,0);
			p = q;
		}
	};

private:
	template<typename U>struct base_Node{
		base_Node() : prev(NULL),next(NULL),element(){};
		base_Node(const base_Node<U> &obj) : prev(obj.prev) , next(obj.prev) , element(obj.element){};
		base_Node(const U &obj) : prev(NULL),next(NULL),element(obj){};
		base_Node<U> *prev;
		base_Node<U> *next;
		U element;
	};
	typedef base_Node<T> _Node;
	_Node *m_head,*m_tail;
	Alloc::rebind<_Node>::other m_nodealloc;
	Alloc m_alloc;
	size_type m_size;


public:

	template<typename U,typename NodeT>struct base_iterator{
		base_iterator() : m_data(NULL),m_nextnull(0),m_prevnull(0){};
		base_iterator(base_Node<NodeT> * const &data) : m_data(data),m_nextnull(0),m_prevnull(0){};
		~base_iterator(){};

		U &operator*(){static U rv;if(m_data)return m_data->element;else return rv;}

		base_iterator<U,NodeT> &operator++(){
			if(m_data == NULL)return *this;
			else if(m_prevnull > 0)--m_prevnull;
			else if(m_data->next)m_data = m_data->next;
			else ++m_nextnull;
			return *this;
		}
		base_iterator<U,NodeT> operator++(int){
			base_iterator<U,NodeT> rv = *this;
			operator++();
			return rv;
		}
		base_iterator<U,NodeT> &operator--(){
			if(m_data == NULL)return *this;
			else if(m_nextnull > 0)--m_nextnull;
			else if(m_data->prev)m_data = m_data->prev;
			else ++m_prevnull;
			return *this;
		}
		base_iterator<U,NodeT> operator--(int){
			base_iterator<U,NodeT> rv = *this;
			operator--();
			return rv;
		}

		bool operator!=(const base_iterator<U,NodeT> &obj) const{return (m_data != obj.m_data);}
		bool operator==(const base_iterator<U,NodeT> &obj) const{return (m_data == obj.m_data);}

		base_iterator<U,NodeT> &operator=(const base_iterator<U,NodeT> &obj){m_data = obj.m_data;return *this;}

		U *operator->() const{return &(m_data->element);}

		base_Node<NodeT> *m_data;
		int m_nextnull,m_prevnull;
	};
	typedef base_iterator<T,T> iterator;
	typedef base_iterator<const T,T> const_iterator;

	iterator begin(){return iterator(m_head);}
	const iterator begin() const{return iterator(m_head);}
//	const_iterator begin() const{return const_iterator(m_head);}

	iterator end(){return iterator(m_tail);}
	const iterator end() const{return iterator(m_tail);}
//	const_iterator end() const{return const_iterator(m_tail);}

	size_type size() const{return m_size;}
	bool empty() const{return (m_size == 0);}

	void insert(iterator pos,const_iterator first,const_iterator last){
		for(;first != last ; ++first)insert(pos,*first);
	}

	iterator insert(iterator pos,const T &ins){
		_Node *New = m_nodealloc.allocate(1);
		m_nodealloc.construct(New,ins);
		New->prev = pos.m_data->prev;
		New->next = pos.m_data;
		if(pos.m_data->prev)pos.m_data->prev->next = New;
		pos.m_data->prev = New;
		if(pos.m_data == m_head)m_head = New;
		++m_size;
		return iterator(New);
	}

	iterator erase(iterator pos){
		if(pos.m_data->next == NULL)return pos;
		_Node *next = pos.m_data->next;
		if(pos.m_data->prev)pos.m_data->prev->next = next;
		next->prev = pos.m_data->prev;
		if(pos.m_data->prev == NULL)m_head = pos.m_data->next;
		m_nodealloc.destroy(pos.m_data);
		m_nodealloc.deallocate(pos.m_data,0);
		--m_size;
		return iterator(next);
	}

	iterator erase(iterator first,iterator last){
		while(first != last)first = erase(first);
		return last;
	}

	void clear(){erase(begin(),end());}

	void push_back(const T &obj){insert(m_tail,obj);}
	void pop_back(){iterator ite = end();--ite;erase(ite);}

};
}//namespace ablib

#endif //__ABLIB_LIST_H__
