#include "ablib.h"
#include "string.h"
#include "SmallBuffer.h"
#include "KRSearch.h"
#include "BMSearch.h"
#include <mbstring.h>

namespace ablib{
	basic_string<char>::size_type basic_string<char>::find(value_type ch, size_type off) const {
		auto p = ::_mbschr((const unsigned char*)(c_str() + off), ch);
		if (p == nullptr)return npos;
		else return (const char*)p - c_str();
	}
	basic_string<char>::size_type basic_string<char>::find(const value_type *ch, size_type off)const {
		auto p = ::_mbsstr((const unsigned char*)(c_str() + off), (const unsigned char*)(ch));
		if (p == nullptr)return npos;
		else return (const char*)p - c_str();
	}
	basic_string<char>::size_type basic_string<char>::find(const value_type *ptr, size_type off, size_type count) const {
		auto searched = basic_string<char>(ptr, count);
		auto p = ::_mbsstr((const unsigned char*)(c_str() + off), (const unsigned char*)searched.c_str());
		if (p == nullptr)return npos;
		else return (const char*)p - c_str();
	}
	basic_string<char>::size_type basic_string<char>::find(const basic_string<char>& str, size_type off) const {
		return find(str.c_str(), off);
	}
	basic_string<wchar_t>::size_type basic_string<wchar_t>::find(value_type ch, size_type off) const {
		return std::basic_string<wchar_t>::find(ch, off);
	}
	basic_string<wchar_t>::size_type basic_string<wchar_t>::find(const value_type *ch, size_type off)const {
		return std::basic_string<wchar_t>::find(ch, off);
	}
	basic_string<wchar_t>::size_type basic_string<wchar_t>::find(const value_type *ptr, size_type off, size_type count) const {
		return std::basic_string<wchar_t>::find(ptr, off, count);
	}
	basic_string<wchar_t>::size_type basic_string<wchar_t>::find(const basic_string<wchar_t>& str, size_type off) const {
		return std::basic_string<wchar_t>::find(str, off);
	}

	/*
	basic_string<char>::size_type basic_string<char>::rfind(value_type ch, size_type off) const {
		auto p = ::_mbsrchr((const unsigned char*)(c_str() + off), ch);
		if (p == nullptr)return npos;
		else return (const char*)p - c_str();
	}
	basic_string<char>::size_type basic_string<char>::rfind(const value_type *ch, size_type off)const {
		auto p = ::str((const unsigned char*)(c_str() + off), (const unsigned char*)(ch));
		if (p == nullptr)return npos;
		else return (const char*)p - c_str();
	}
	basic_string<char>::size_type basic_string<char>::rfind(const value_type *ptr, size_type off, size_type count) const {
		auto searched = basic_string<char>(ptr, count);
		auto p = ::_mbsstr((const unsigned char*)(c_str() + off), (const unsigned char*)searched.c_str());
		if (p == nullptr)return npos;
		else return (const char*)p - c_str();
	}
	basic_string<char>::size_type basic_string<char>::rfind(const basic_string<char>& str, size_type off) const {
		return rfind(str.c_str(), off);
	}
	basic_string<wchar_t>::size_type basic_string<wchar_t>::rfind(value_type ch, size_type off) const {
		return std::basic_string<wchar_t>::rfind(ch, off);
	}
	basic_string<wchar_t>::size_type basic_string<wchar_t>::rfind(const value_type *ch, size_type off)const {
		return std::basic_string<wchar_t>::rfind(ch, off);
	}
	basic_string<wchar_t>::size_type basic_string<wchar_t>::rfind(const value_type *ptr, size_type off, size_type count) const {
		return std::basic_string<wchar_t>::rfind(ptr, off, count);
	}
	basic_string<wchar_t>::size_type basic_string<wchar_t>::rfind(const basic_string<wchar_t>& str, size_type off) const {
		return std::basic_string<wchar_t>::rfind(str, off);
	}*/


const basic_string<char> &basic_string<char>::tolower(){
	char *buf = new char[length() + 1];
	::lstrcpyA(buf,c_str());
	::CharLowerBuffA(buf,(DWORD)length());
	assign(buf);
	delete[] buf;
	return *this;
}

const basic_string<wchar_t> &basic_string<wchar_t>::tolower(){
	wchar_t *buf = new wchar_t[length() + 1];
	::lstrcpyW(buf,c_str());
	::CharLowerBuffW(buf,(DWORD)length());
	assign(buf);
	delete[] buf;
	return *this;
}

const basic_string<char> &basic_string<char>::toupper(){
	char *buf = new char[length() + 1];
	::lstrcpyA(buf,c_str());
	::CharUpperBuffA(buf,(DWORD)length());
	assign(buf);
	delete[] buf;
	return *this;
}

const basic_string<wchar_t> &basic_string<wchar_t>::toupper(){
	wchar_t *buf = new wchar_t[length() + 1];
	::lstrcpyW(buf,c_str());
	::CharUpperBuffW(buf,(DWORD)length());
	assign(buf);
	delete[] buf;
	return *this;
}

#ifndef _CRT_NON_CONFORMING_SWPRINTFS
#define _CRT_NON_CONFORMING_SWPRINTFS
#endif
#ifndef _CRT_SECURE_NO_WARNINGS
#define _CRT_SECURE_NO_WARNINGS
#endif
basic_string<char> basic_string<char>::sprintf(const char* fmt,...){
	basic_string<char> rv;
	ablib::SmallBuffer<char> buf;
	va_list args;
	va_start(args,fmt);
	buf.Alloc(::_vscprintf (fmt,args));
	::vsprintf(buf.buf,fmt,args);
	va_end(args);
	rv = buf.buf;
	return rv;
}

basic_string<wchar_t> basic_string<wchar_t>::sprintf(const wchar_t* fmt,...){
	basic_string<wchar_t> rv;
	ablib::SmallBuffer<wchar_t> buf;
	va_list args;
	va_start(args,fmt);
	buf.Alloc(::_vscwprintf(fmt,args));
	::vswprintf(buf.buf, buf.size, fmt, args);
	va_end(args);
	rv = buf.buf;
	return rv;
}


namespace TransCode{

basic_string<char> Trans(const basic_string<wchar_t> &str){
	ablib::SmallBuffer<char> buf;
	DWORD size = WideCharToMultiByte(CP_ACP,0,str.c_str(),(int)str.length(),NULL,0,NULL,NULL);
	buf.Alloc(size + 1);
	size = ::WideCharToMultiByte(CP_ACP,0,str.c_str(),(int)str.length(),buf.buf,(int)buf.size,NULL,NULL);
	buf.buf[size] = '\0';
	return basic_string<char>(buf.buf);
}

basic_string<wchar_t> Trans(const basic_string<char> &str){
	SmallBuffer<wchar_t> buf;
	DWORD size = ::MultiByteToWideChar(CP_ACP,MB_PRECOMPOSED,str.c_str(),(int)str.length(),NULL,0);
	buf.Alloc(size + 1);
	size = ::MultiByteToWideChar(CP_ACP,MB_PRECOMPOSED,str.c_str(),(int)str.length(),buf.buf,(int)buf.size);
	buf.buf[size] = L'\0';
	return basic_string<wchar_t>(buf.buf);
}
}

}//namespace ablib
