#include "edit.h"
#include "EditPos.h"
#include "../ablib/IsKanji.h"

namespace ablib{
namespace window{
namespace edit{

void EditPos::BeginSel(POS begin){
	if(m_beginsel)return;
	CheckPos(begin);
	if(!m_selecting)m_startsel = begin;
	m_seltop = m_selend = begin;
	m_beginsel = true;
	m_selecting = true;
}

bool EditPos::SetPos(POS pos){
	bool rv = CheckPos(pos);
	m_pos = pos;
	m_xy = m_buf->POSToXY(pos);
	m_rx = m_xy.x;
	if(m_beginsel){
		if(pos > m_startsel){
			m_seltop = m_startsel;
			m_selend = pos;
		}else{
			m_seltop = pos;
			m_selend = m_startsel;
		}
	}
	return !rv;
}

void EditPos::EndSel(POS end){
	if(!m_beginsel)return;
	CheckPos(end);
	m_beginsel = false;
	if(m_startsel < end){
		m_seltop = m_startsel;
		m_selend = end;
	}else{
		m_seltop = end;
		m_selend =m_startsel;
	}
	if(m_seltop == m_selend)m_selecting = false;
}

bool EditPos::CheckPos(POS &pos){
	bool rv = false;
	if(pos.lineno < 0){
		pos.lineno = 0;
		rv = true;
	}
	if(pos.column < 0){
		pos.column = 0;
		rv = true;
	}
	if(pos.lineno > m_buf->GetLineCount()){
		pos.lineno = m_buf->GetLineCount();
		rv = true;
	}
	const TCHAR *str = m_buf->GetText(pos.lineno);
	if(str == NULL){
		pos.column = 0;
		rv = true;
	}else if(::lstrlen(str) < pos.column){
		pos.column = ::lstrlen(str);
		rv = true;
#ifndef UNICODE
	}else if(ablib::IsKanji2(str,pos.column)){
		--pos.column;
		rv = true;
#endif
	}
	return rv;
}

bool EditPos::Right(int n){
	ASSERT(n >= 0);
	if(n == 1){
		POS pos = m_pos;
		const TCHAR *text = m_buf->GetText(pos.lineno);
		if(text == NULL)return false;
		if(text[pos.column] == '\0' && pos.lineno == m_buf->GetLineCount() - 1){m_rx = 0;return false;}
#ifdef UNICODE
		++pos.column;
#else
		if(ablib::IsKanji(text[pos.column]))pos.column += 2;
		else ++pos.column;
#endif
		if(pos.column > ::lstrlen(text)){++pos.lineno;pos.column = 0;}
		SetPos(pos);
		return true;
	}else{
		POS pos = m_pos;
		pos.column += n;
		while(true){
			const TCHAR *text = m_buf->GetText(pos.lineno);
			if(text == NULL)return false;
			int len = ::lstrlen(text);
			if(pos.column < len){
#ifndef UNCODE
				if(ablib::IsKanji2(text,pos.column))--pos.column;
#endif
				SetPos(pos);
				return true;
			}
			++pos.lineno;
			if(pos.lineno >= m_buf->GetLineCount())return false;
			pos.column -= len;
		}
	}
}

bool EditPos::Left(int n){
	ASSERT(n >= 0);
	const TCHAR *text = _T("");
	POS pos = m_pos;
	pos.column -= n;
	while(true){
		if(pos.column >= 0){
			text = m_buf->GetText(pos.lineno);
			if(text == NULL)return false;
#ifndef UNICODE
			if(ablib::IsKanji2(text,pos.column))--pos.column;
#endif
			SetPos(pos);
			return true;
		}
		if(pos.lineno == 0){m_rx = 0;return false;}
		--pos.lineno;
		text = m_buf->GetText(pos.lineno);
		if(text == NULL)return false;
		pos.column += ::lstrlen(text) + 1;
	}
}

bool EditPos::Up(int n){
	ASSERT(n >= 0);
	XY xy = m_xy;
	xy.y -= n;
	if(xy.y < 0)xy.y = 0;
	LayoutData ld = m_buf->GetLayoutData(xy.y);
	int rx = m_rx;
	xy.x = Min(m_rx,m_buf->GetLineLen(xy.y));
	POS pos = m_buf->XYToPOS(xy);
	const TCHAR *text = m_buf->GetText(pos.lineno);
#ifndef UNICODE
	if(ablib::IsKanji2(text + ld.offset,pos.column - ld.offset))--pos.column;
#endif
	SetPos(pos);
	m_rx = rx;
	return true;
}

bool EditPos::Down(int n){
	ASSERT(n >= 0);
	XY xy = m_xy;
	xy.y += n;
	if(xy.y > m_buf->GetLayoutLineCount() - 1)xy.y = m_buf->GetLayoutLineCount() - 1;
	LayoutData ld = m_buf->GetLayoutData(xy.y);
	int rx = m_rx;
	xy.x = Min(m_rx,m_buf->GetLineLen(xy.y));
	POS pos = m_buf->XYToPOS(xy);
	const TCHAR *text = m_buf->GetText(pos.lineno);
#ifndef UNICODE
	if(ablib::IsKanji2(text + ld.offset,pos.column - ld.offset))--pos.column;
#endif
	SetPos(pos);
	m_rx = rx;
	return true;
}

void EditPos::CorrectXY(XY &xy){
	if(xy.y < 0)xy.y = 0;
	else if(xy.y >= m_buf->GetLayoutLineCount())xy.y = m_buf->GetLayoutLineCount() - 1;
	if(xy.x < 0)xy.x = 0;
	else{
		LayoutData ld = m_buf->GetLayoutData(xy.y);
		POS pos = POS(ld.line,ld.offset + ld.length - 1);
		int x = m_buf->POSToXY(pos).x + 1;
		if(xy.x > x)xy.x = x;
		int line = m_buf->YToLine(xy.y);
		const TCHAR *str = m_buf->GetText(line);
#ifndef UNICODE
		if(ablib::IsKanji2(str + ld.offset,xy.x))--xy.x;
#endif
	}
}

void EditPos::SetSel(POS start,POS end){
	if(start < end){
		m_seltop = start;
		m_selend = end;
	}else{
		m_seltop = end;
		m_selend = start;
	}
	m_selecting = true;
}

}//namespace edit
}//namespace window
}//namespace ablib
