#include "edit.h"
#include <windowsx.h>
#include "edit_const.h"
#include "editwnd.h"
#include "command.h"

namespace ablib{
namespace window{

//܂ł傤D
using namespace edit;

EditWnd::EditWnd(HINSTANCE appint) :
	m_search(&m_buf),
	m_pos(&m_buf),
	m_isdraw(true),
	m_scrollpage(0),
	m_onescroll(0),
	m_scrollmax(0),
	BaseWnd(EditWindowClassName,WS_CHILD | WS_BORDER | WS_VISIBLE | WS_VSCROLL | WS_HSCROLL,appint)
	{
	SetFont();
	EditConfig c;
	c.tabsize = 4;
	c.undolimit = 100;
	c.wrapsize = 80;
	c.fontsize = 16;
	c.fontname = _T("lr SVbN");

	COLORREF window = ::GetSysColor(COLOR_WINDOW);
	COLORREF wtext = ::GetSysColor(COLOR_WINDOWTEXT);
	COLORREF hwindow = ::GetSysColor(COLOR_HIGHLIGHT);
	COLORREF hwtext = ::GetSysColor(COLOR_HIGHLIGHTTEXT);

	c.NormalColor = Color(rgb(128,200,128),rgb(128,200,128),rgb(wtext));
	c.NormalBackColor = Color(rgb(window),rgb(window),rgb(window));
	c.SelectColor = Color(rgb(0),rgb(255,255,255),rgb(hwtext));
	c.SelectBackColor = Color(rgb(0),rgb(100,180,100),rgb(hwindow));

	SetConfig(c);

	m_set.overwrite = false;
	m_set.readonly = false;
}


bool EditWnd::OnCreate(Window *wnd){
	m_fontdata = wnd->GetFontData();
	Rect rc = wnd->GetRect();
	wnd->GetVScrollBar()->SetPage((rc.bottom - rc.top - TOP_SPACE) / TextHeight());
	wnd->GetHScrollBar()->SetPage((rc.right - rc.left - LEFT_SPACE)/ m_fontdata.avewidth);

	wnd->GetVScrollBar()->SetPos(0);
	wnd->GetHScrollBar()->SetPos(0);

//	m_pos.Init(hwnd);
//	m_pos.SetFont(m_textmetric);
	wnd->GetIME()->SetFont(ablib::window::FontData(m_config.fontname.c_str(),m_config.fontsize));
	SetFont();

	return true;
}

bool EditWnd::OnSize(Window *wnd){
	SetScroll();
	if(m_pos.GetXY().y > wnd->GetVScrollBar()->GetPos() + wnd->GetVScrollBar()->GetPage())
		wnd->GetVScrollBar()->SetPos(m_pos.GetXY().y + wnd->GetVScrollBar()->GetPage());
	return true;
}

bool EditWnd::OnSetFocus(Window *wnd,Window *old){
	SetFont();
	return true;
}

bool EditWnd::OnKillFocus(Window *wnd,Window *New){
	wnd->GetCaret()->Destroy();
	return true;
}

void EditWnd::SetFont(){
	if(GetWindow()){
		GetWindow()->GetIME()->SetFont(ablib::window::FontData(m_config.fontname.c_str(),m_config.fontsize));
		GetWindow()->SetFont(ablib::window::FontData(m_config.fontname.c_str(),m_config.fontsize));
		m_fontdata = GetWindow()->GetFontData();
		SetCaret(false);
		SetCaret(true);
//		SetScrollBar(GetWindow());
//		m_pos.SetFont(m_textmetric);
	}
}

void EditWnd::SetConfig(const EditConfig &c){
	m_config = c;
	if(m_config.tabsize <= 0)m_config.tabsize = 1;
	SetFont();
	m_buf.SetTabSize(m_config.tabsize);
	m_buf.SetMaxX(m_config.wrapsize);
	if(m_config.wrapsize > 0)m_buf.SetDoLayout(true);
	else m_buf.SetDoLayout(false);
	m_undo.SetMax(m_config.undolimit);
}

bool EditWnd::AddHilight(const ablib::string &keyword,int opt,int hilightpos,rgb color,rgb back){
	if(!m_buf.AddHilight(keyword.c_str(),opt,hilightpos))return false;
	m_hilightcolor.push_back(HilightColor(color,back));
	Invalidate(POS(0,0),POS(-1,-1));
	return true;
}

//shiftȂǂ獡͈͑IׂvZD
//}EX͍lȂ
//Ăǂ΂̂悭킩ȂD
void EditWnd::CheckSelect(){
	bool shift = (::GetKeyState(VK_SHIFT) < 0);
	if(shift && !m_pos.IsBeginSel())beginsel();
	else if(!shift && selecting()){endsel();escape();}
}

void EditWnd::SetCaret(bool b){
	if(!GetWindow())return;
	if(b){
		// 肦[
		if(GetWindow()->GetCaret()->IsCreated())GetWindow()->GetCaret()->Hide();
		else GetWindow()->GetCaret()->Create(m_set.overwrite ? m_fontdata.avewidth : 1,TextHeight());
		int he = TextHeight();
		XY xy = m_pos.GetXY();
		POS pos = m_pos.GetPos();
		Point pt(
			(m_pos.GetXY().x - GetWindow()->GetHScrollBar()->GetPos()) * m_fontdata.avewidth + LEFT_SPACE,
			(m_pos.GetXY().y - GetWindow()->GetVScrollBar()->GetPos()) * TextHeight() + TOP_SPACE
			);
		GetWindow()->GetCaret()->SetPos(pt);
		GetWindow()->GetCaret()->Show();
		GetWindow()->GetIME()->SetPos(pt);
	}else GetWindow()->GetCaret()->Destroy();
}

void EditWnd::SetScroll(){
	if(m_isdraw && GetWindow()){
		int oldvpos = GetWindow()->GetVScrollBar()->GetPos();
		int oldhpos = GetWindow()->GetHScrollBar()->GetPos();

		Rect rc = GetWindow()->GetRect();
		GetWindow()->GetVScrollBar()->SetPage((rc.bottom - rc.top - TOP_SPACE) / TextHeight());
		GetWindow()->GetVScrollBar()->SetRange(0,m_buf.GetLayoutLineCount());

		GetWindow()->GetHScrollBar()->SetPage((rc.right - rc.left - LEFT_SPACE) / m_fontdata.avewidth);
		GetWindow()->GetHScrollBar()->SetRange(0,m_buf.GetMaxLen());

		int scpos = GetWindow()->GetVScrollBar()->GetPos();
		int y = m_pos.GetXY().y;
		int page = GetWindow()->GetVScrollBar()->GetPage();
		if(y >= scpos + page)GetWindow()->GetVScrollBar()->SetPos(y - page + 1);
		else if(y <= scpos)GetWindow()->GetVScrollBar()->SetPos(y);

		scpos = GetWindow()->GetHScrollBar()->GetPos();
		int x = m_pos.GetXY().x;
		page = GetWindow()->GetHScrollBar()->GetPage();
		if(x >= scpos + page)GetWindow()->GetHScrollBar()->SetPos(x - page + 1);
		else if(x <= scpos)GetWindow()->GetHScrollBar()->SetPos(x);

		if(oldvpos != GetWindow()->GetVScrollBar()->GetPos() || oldhpos != GetWindow()->GetHScrollBar()->GetPos()){
			GetWindow()->ScrollWindow(
				(oldhpos - GetWindow()->GetHScrollBar()->GetPos()) * m_fontdata.avewidth,
				(oldvpos - GetWindow()->GetVScrollBar()->GetPos()) * TextHeight()
				);
			if(oldvpos != GetWindow()->GetVScrollBar()->GetPos()){
				Rect rc = GetWindow()->GetRect();
				if(oldvpos > GetWindow()->GetVScrollBar()->GetPos()){
					rc.top = TextHeight() * (oldvpos - GetWindow()->GetVScrollBar()->GetPos());
					rc.bottom = rc.top + TOP_SPACE;
				}else{
					rc.top = 0;
					rc.bottom = TOP_SPACE;
				}
				GetWindow()->Invalidate(rc);
			}
			if(oldhpos != GetWindow()->GetHScrollBar()->GetPos()){
				Rect rc = GetWindow()->GetRect();
				if(oldhpos > GetWindow()->GetHScrollBar()->GetPos()){
					rc.left = m_fontdata.avewidth * (oldhpos - GetWindow()->GetHScrollBar()->GetPos());
					rc.right = rc.left + LEFT_SPACE;
				}else{
					rc.left = 0;
					rc.right = LEFT_SPACE;
				}
				GetWindow()->Invalidate(rc);
			}
		}
	}
}

void EditWnd::InvalidateSel(POS oldtop,POS oldend){
	POS start;
	POS end;
	if(oldend == m_pos.SelEnd()){
		if(oldtop > m_pos.SelTop()){
			start = m_pos.SelTop();
			end = oldtop;
		}else{
			start = oldtop;
			end = m_pos.SelTop();
		}
	}else if(oldtop == m_pos.SelTop()){
		if(oldend > m_pos.SelEnd()){
			start = m_pos.SelEnd();
			end = oldend;
		}else{
			start = oldend;
			end = m_pos.SelEnd();
		}
	}else{
		start = Min(m_pos.SelTop(),oldtop);
		end = Max(m_pos.SelEnd(),oldend);
	}
	Invalidate(start,end,m_buf.GetText(start.lineno));
/*
	POS start = Min(oldtop,m_pos.SelTop());
	POS end = Max(oldend,m_pos.SelEnd());
	Invalidate(start,end,m_buf.GetText(start.lineno));
*/
}

}//namespace window
}//namsepace ablib

