/** 
* @file EditWnd.h
* @brief GfB^pʂ̒
* @author Abe Noriyuki
*/

#ifndef __ABLIB_WINDOW_EDITWND_H__
#define __ABLIB_WINDOW_EDITWND_H__


//#include "../ablib/vector.h"
#include <vector>
#include "../ablib/SmallVector.h"
#include "../ablib/function.h"
#include "../window/window.h"
#include "../window/Graphic.h"
#include "../window/basewnd.h"
#include "EditBuf.h"
#include "EditPos.h"
#include "Command.h"
#include "RedoUndo.h"
#include "search.h"

namespace ablib{
namespace window{
namespace edit{

/**
* GfBb^̃EBhÊ悤Ȃ̂񋟂NXłD
* Gۂ̃}N݂ȖOĂ܂DS݊ł͂ȂǁD
* 
* ʂȏ͌pOn***ĂD
* ɂ̃NXɕĎD̃NX͎ɕ`SD
* ƊOւ̃C^[tF[XJD̃NXeditOԓɂ܂D
*/

class EditWnd : public BaseWnd{
//EBhE{֐
public:
	EditWnd(HINSTANCE appint);
	~EditWnd(){};
	Window *Create(Wnd *parent){
		return BaseWnd::Create(parent,::LoadCursor(NULL,IDC_IBEAM));
	}

	friend edit::InsertCommand;
	friend edit::DeleteCommand;
	friend edit::ReplaceCommand;

private:
	HINSTANCE m_appint;

	//`
public:
	struct TabData{
		int pos;
		int tabs;
		TabData(int p,int t) : pos(p),tabs(t){};
		TabData(){};
	};

private:
	void DrawLine(Window *wnd,Graphic &g,int line,ablib::string &str,SmallVector<int> &offsets,int start,int end);
	void DrawLineEnd(Window *wnd,Graphic &g,int endx,int endy,bool sel);
	void Invalidate(edit::POS start,edit::POS end,const TCHAR *oldtopstr = NULL);
	bool m_isdraw;

//Cxgւ̉
protected:
	bool OnCreate(Window *wnd);
//	bool OnDestroy(Window *wnd);
	bool OnPaint(Window *wnd);
	bool OnSize(Window *wnd);
	bool OnSetFocus(Window *wnd,Window *old);
	bool OnKillFocus(Window *wnd,Window *New);
	bool OnMouseMove(Window *wnd,Mouse *mos,int keyFlags);
	bool OnLButtonDown(Window *wnd,Mouse *mos,bool isDClick);
	bool OnLButtonUp(Window *wnd,Mouse *mos);
//	bool OnRButtonDown(Window *wnd,Mouse *mos,bool isDClick);
//	bool OnRButtonUp(Window *wnd,Mouse *mos);
//	bool OnMButtonDown(Window *wnd,Mouse *mos,bool isDClick);
//	bool OnMButtonUp(Window *wnd,Mouse *mos);
	bool OnMouseWheel(Window *wnd,Mouse *mos,int delta);
	bool OnKey(Window *wnd,bool isDown,UINT vk,int repeat);
	bool OnChar(Window *wnd,TCHAR ch,int repeat);

	bool OnIME(Window *wnd);

//L[
public:
	void CheckSelect();//J[\ړɎgp
	void InvalidateSel(edit::POS oldtop,edit::POS oldend);//helper

private:
	edit::EditBuf m_buf;

//tHg֐
private:
	void SetFont();

private:
	FontData m_fontdata;
	int TextHeight(){return m_fontdata.height;}

//XN[o[
private:
	int m_scrollpage;
	int m_onescroll;
	int m_scrollmax;

	bool OnHScroll(Window *wnd,int dx);
	bool OnVScroll(Window *wnd,int dy);
	void SetScroll();

//caretiLbgHJbgHj
private:
	edit::EditPos m_pos;
	void SetCaret(bool b);

//undoredo
private:
	edit::RedoUndo m_undo;

//
private:
	edit::Search m_search;
	ablib::string m_replacestr;

public:
	enum SearchOption{
		word = 1,
		casesense = 2,
		replace = 4,
		ask = 8,
		regular = 16,
		fuzzy = 32,
		inselect = 64,
	};

//ݒ
public:
	struct Color{
		Color(rgb e,rgb r,rgb n) : eof(e),Return(r),normal(n){}
		Color() : eof(0,0,0),Return(0,0,0),normal(0,0,0){}
		rgb eof;
		rgb Return;
		rgb normal;
	};

	//ݒp\́i̕ӂ͌ōlKvC܂Dj
	struct EditConfig{
		Color NormalColor;
		Color SelectColor;
		Color NormalBackColor;
		Color SelectBackColor;

		int tabsize;
		int wrapsize;

		int undolimit;
	
		ablib::string fontname;
		int fontsize;
	};

	void SetConfig(const EditConfig &c);
	EditConfig GetConfig() const{return m_config;}

	bool AddHilight(const ablib::string &keyword,int opt,int hilightpos,rgb color,rgb back);
	void ClearHilight(){m_buf.HilightClear();Invalidate(edit::POS(0,0),edit::POS(-1,-1));}

private:
	EditConfig m_config;
	struct HilightColor{
		rgb text;
		rgb back;
		HilightColor(rgb t,rgb b) : text(t),back(b){}
		HilightColor() : text(0,0,0),back(0,0,0){}
	};
	SmallVector<HilightColor> m_hilightcolor;


	struct EditSet{
		bool readonly;
		bool overwrite;
	}m_set;

//ȉO瑀삷邽ߗpD
public:
	//Gۃ}NɂȂD
	const TCHAR *GetText(int line){return m_buf.GetText(line);}
	void RemoveAllText(){
		m_pos.SetPos(edit::POS(0,0));m_pos.SetSel(edit::POS(0,0),edit::POS(0,0));
		m_buf.RemoveAll();SetCaret(true);if(GetWindow())GetWindow()->Invalidate();
	}
	void ResetUpdated(){m_undo.ResetUpdated();}
	void ClearUndoHistroty(){m_undo.Clear();}

public:
	//֐
	const TCHAR *gettext(int x1,int y1,int x2,int y2);

public:
	//J[\ړ
	bool movetolineno(int column,int lineno);
	bool moveto(int x,int y);
	bool down(int n = 1);
	bool gofileend();
	bool gofiletop(){return movetolineno(1,1);}
	bool golineend();
	bool golineend2();
	bool golinetop(){return moveto(0,y());}
	bool golinetop2(){return movetolineno(1,lineno());}
	bool left(int n = 1);
	bool up(int n = 1);
	bool nextpage();
	bool prevpage();
	bool right(int n = 1);

	//}n
	bool insert(const ablib::string &str,const TCHAR *sep = _T("\n"));
	bool insertreturn(){return insert(_T("\n"));}
	bool overwrite(const ablib::string &str);
	bool tab();

	//Nbv{[hn
	bool beginsel(){m_pos.BeginSel();SetCaret(true);return true;}
	bool endsel(){m_pos.EndSel();return true;}
	bool copy();
	bool cut();
	bool escape(){if(selecting()){m_pos.Escape();Invalidate(m_pos.SelTop(),m_pos.SelEnd(),m_buf.GetText(m_pos.SelTop().lineno));}return true;}
	bool paste();
	bool selectall(){gofileend();beginsel();gofiletop();endsel();return true;}
	bool selectline(){golineend2();beginsel();golinetop();endsel();return true;}

	//폜n
	bool backspace();
	bool Delete();
	
	//̑ҏWn
	bool undo(){bool rv = m_undo.Undo(this);return rv;}
	bool redo(){bool rv = m_undo.Redo(this);return rv;}

	//n
	bool searchdown(const ablib::string &search,const ablib::string &flag = _T(""));
	bool searchup(const ablib::string &search,const ablib::string &flag = _T(""));
	bool replacedown(const ablib::string &search,const ablib::string &rep,const ablib::string &flag = _T(""));
	bool replaceup(const ablib::string &search,const ablib::string &rep,const ablib::string &flag = _T(""));
	int replaceall(const ablib::string &search,const ablib::string &rep,const ablib::string &flag = _T(""),bool up = false);

	bool finddown();
	bool findup();
	ablib::string getsearch();
	void setsearch(const ablib::string &str,const ablib::string &flag){m_search.SetSearch(str,flag);}
	void setsearch(const ablib::string &str,int flag){m_search.SetSearch(str,flag);}

	//̑R}hn
	void imeswitch(){GetWindow()->GetIME()->SetOpen(!GetWindow()->GetIME()->IsOpen());}
	void overwriteswitch(){SetCaret(false);m_set.overwrite = !overwrite();SetCaret(true);}
	void readonlyswitch(){m_set.readonly = !readonly();}

	//}NiHjn
	void disabledraw();
	void enabledraw();

//L[[h
	//J[\֘A
	int x() const{return m_pos.GetXY().x;}
	int y() const{return m_pos.GetXY().y;}
	int code();
	int column() const{return m_pos.GetPos().column;}
	int lineno() const{return m_pos.GetPos().lineno + 1;}

	//eLXg֘A
	int linecount() const{return m_buf.GetLayoutLineCount();}
	int linecount2() const{return m_buf.GetLineCount();}
	int linelen();
	int linelen2();

	//͈͑I֘A
	bool selecting() const{return m_pos.IsSelecting();}
	int seltopx(){return m_buf.POSToXY(m_pos.SelTop()).x;}
	int seltopy(){return m_buf.POSToXY(m_pos.SelTop()).y;}
	int selendx(){return m_buf.POSToXY(m_pos.SelEnd()).x;}
	int selendy(){return m_buf.POSToXY(m_pos.SelEnd()).y;}

	//eԊ֘A
	bool overwrite() const{return m_set.overwrite;}
	bool updated() const{return m_undo.Updated();}
	bool imestate() const{return GetWindow()->GetIME()->IsOpen();}
	const TCHAR *searchbuffer() const{return m_search.SearchBuffer().c_str();}
	int searchoption() const{return m_search.SearchOption();}
	bool readonly()const {return m_set.readonly;}

	//t@C^CvʁiHj̐ݒ֘A
	const TCHAR *fontname(){return m_config.fontname.c_str();}
	int fontsize(){return TextHeight();}
	int tabcount(){return m_config.tabsize;}
	int width(){return m_config.wrapsize;}

private:
	NOCOPY(EditWnd);
};

}//namespace edit
}//namespace window
}//namespace ablib

#endif //__ABLIB_WINDOW_EDITWND_H__

