#ifndef __ABLIB_SYSTEM_DDESRV_H__
#define __ABLIB_SYSTEM_DDESRV_H__

#include "../ablib/string.h"
#include <functional>
#include "ablib_system.h"
#include <map>

namespace ablib{
namespace system{

class DdeSrv
{
public:
	typedef std::function<bool ()> DDEEXECUTE;
	typedef std::function<bool (TCHAR *)> DDEPOKE;
	typedef std::function<const TCHAR * ()> DDEREQUEST;

	DdeSrv(const TCHAR *service,const TCHAR *topic)
		: m_servicestr(service),m_topicstr(topic),m_ddeinst(0),m_connect(false),m_advice(false){}
	~DdeSrv(){UnInit();}
	bool Init();
	bool UnInit();
	bool PostAdvice(const TCHAR *itemstr,const TCHAR *data);
	void SetDdeExecute(const TCHAR *str,const DDEEXECUTE &exec){m_ddeexec[str] = exec;}
	void DelDdeExecute(const TCHAR *str){m_ddeexec.erase(str);}
	void SetDdePoke(const TCHAR *str,const DDEPOKE &poke){m_ddepoke[str] = poke;}
	void DelDdePoke(const TCHAR *str){m_ddepoke.erase(str);}
	void SetDdeRequest(const TCHAR *str,const DDEREQUEST &req){m_dderequest[str] = req;}
	void DelDdeRequest(const TCHAR *str){m_dderequest.erase(str);}

private:
	bool m_connect,m_advice;
	DWORD m_ddeinst;
	ablib::string m_servicestr,m_topicstr;
	HSZ m_service,m_topic;
	ablib::string m_advicestr;

	std::map<ablib::string,DDEEXECUTE> m_ddeexec;
	std::map<ablib::string,DDEPOKE   > m_ddepoke;
	std::map<ablib::string,DDEREQUEST> m_dderequest;
	static HDDEDATA CALLBACK DdeProc(UINT uType,UINT uFmt,HCONV hconv,HSZ hsz1,HSZ hsz2,HDDEDATA hdata,ULONG_PTR dwData1,ULONG_PTR dwData2);
	static DWORD TransDdeStr(DWORD ddeinst,HSZ hsz,TCHAR *buf,int bufsize){
#ifdef UNICODE
		return ::DdeQueryString(ddeinst,hsz,buf,bufsize,CP_WINUNICODE);
#else
		return ::DdeQueryString(ddeinst,hsz,buf,bufsize,CP_WINANSI);
#endif
	}

	static std::map<HSZ,DdeSrv*> m_topic2pt;
	static std::map<HCONV,DdeSrv*> m_hconv2pt;

};

}//namespace system
}//namespace ablib

#endif //__ABLIB_SYSTEM_DDESRV_H__

