#include "ablib_system.h"
#include "file.h"
#include "FindFile.h"

namespace ablib{
namespace system{

DWORD File::GetSize(const TCHAR *file){
	HANDLE h = ::CreateFile(file,0,FILE_SHARE_READ,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
	if(h == INVALID_HANDLE_VALUE){SetLastError();return 0;}
	DWORD rv = ::GetFileSize(h,NULL);
	::CloseHandle(h);
	return rv;
}

Time File::GetCreationTime(const TCHAR *file){
	HANDLE han = ::CreateFile(file,0,0,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
	if(han == INVALID_HANDLE_VALUE){SetLastError();return Time();}
	FILETIME ft;
	if(!::GetFileTime(han,&ft,NULL,NULL)){
		::CloseHandle(han);
		return Time();
	}
	::CloseHandle(han);
	SYSTEMTIME st;
	if(::FileTimeToSystemTime(&ft,&st))return Time(st);
	else{SetLastError();return Time();}
}
	
Time File::GetLastAccessTime(const TCHAR *file){
	HANDLE han = ::CreateFile(file,0,0,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
	if(han == INVALID_HANDLE_VALUE){SetLastError();return Time();}
	FILETIME ft;
	if(!::GetFileTime(han,NULL,&ft,NULL)){
		SetLastError();
		::CloseHandle(han);
		return Time();
	}
	::CloseHandle(han);
	SYSTEMTIME st;
	if(::FileTimeToSystemTime(&ft,&st))return Time(st);
	else{SetLastError();return Time();}
}

Time File::GetLastWriteTime(const TCHAR *file){
	//HANDLE han = ::CreateFile(file,GENERIC_READ,0,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
	HANDLE han = ::CreateFile(file, 0, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if(han == INVALID_HANDLE_VALUE){
		SetLastError();
		::CloseHandle(han);
		return Time();
	}
	FILETIME ft;
	if(!::GetFileTime(han,NULL,NULL,&ft)){CloseHandle(han);SetLastError();return Time();}
	::CloseHandle(han);
	SYSTEMTIME st;
	if(::FileTimeToSystemTime(&ft,&st))return Time(st);
	else{SetLastError();return Time();}
}

bool File::SetCreationTime(const TCHAR *file,const Time &t){
	FILETIME ft;
	if(!::SystemTimeToFileTime(&(t.m_st),&ft)){SetLastError();return false;}
	HANDLE han = ::CreateFile(file,GENERIC_WRITE,0,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
	if(han == INVALID_HANDLE_VALUE){SetLastError();return false;}
	if(!::SetFileTime(han,&ft,NULL,NULL)){
		SetLastError();
		::CloseHandle(han);
		return false;
	}
	::CloseHandle(han);
	return true;
}

bool File::SetLastAccessTime(const TCHAR *file,const Time &t){
	FILETIME ft;
	if(!::SystemTimeToFileTime(&(t.m_st),&ft)){SetLastError();return false;}
	HANDLE han = ::CreateFile(file,GENERIC_WRITE,0,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
	if(han == INVALID_HANDLE_VALUE){SetLastError();return false;}
	if(!::SetFileTime(han,NULL,&ft,NULL)){
		SetLastError();
		::CloseHandle(han);
		return false;
	}
	::CloseHandle(han);
	return true;
}

bool File::SetLastWriteTime(const TCHAR *file,const Time &t){
	FILETIME ft;
	if(!::SystemTimeToFileTime(&(t.m_st),&ft)){SetLastError();return false;}
	HANDLE han = ::CreateFile(file,GENERIC_WRITE,0,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
	if(han == INVALID_HANDLE_VALUE){SetLastError();return false;}
	if(!::SetFileTime(han,NULL,NULL,&ft)){
		SetLastError();
		::CloseHandle(han);
		return false;
	}
	::CloseHandle(han);
	return true;
}


bool File::DeleteFiles(const ablib::string &dir){
	bool rv = true;
	WIN32_FIND_DATA wfd;
	ablib::system::FindFile ff;
	ablib::string pattern = dir;
	if(pattern != _T("") && pattern[pattern.length() - 1] != '\\')pattern += _T("\\");
	ablib::string d = pattern;
	pattern += _T("*.*");
	if(ff.Begin(pattern)){
		while(ff.Next(&wfd)){
			if(::lstrcmp(wfd.cFileName,_T(".")) == 0 || ::lstrcmp(wfd.cFileName,_T("..")) == 0)continue;
			d = dir;
			d += wfd.cFileName;
			if(wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY){
				d += _T("\\");
				rv = (rv && File::DeleteFiles(d.c_str()));
			}else{
				rv = (rv && (::DeleteFile(d.c_str()) ? true : false));
			}
		}
		ff.Close();
	}

	d = dir;
	d.erase(d.length() - 1);
	rv = (rv && (::RemoveDirectory(d.c_str()) ? true : false));
	return rv;
}



}//namespace system
}//namespace ablib