#include "ablib_system.h"
#include <tlhelp32.h>
#include "process.h"

namespace ablib{
namespace system{

bool Process::Start(){
	if(!HasExited()){SetLastError(ERROR_SUCCESS);return false;}
	Close();

	SECURITY_ATTRIBUTES sa;
	::ZeroMemory(&sa,sizeof(SECURITY_ATTRIBUTES));
	sa.nLength = sizeof(SECURITY_ATTRIBUTES);
	sa.bInheritHandle = TRUE;
	sa.lpSecurityDescriptor = NULL;

	STARTUPINFO si;
	::ZeroMemory(&si,sizeof(STARTUPINFO));
	si.cb = sizeof(STARTUPINFO);

	HANDLE AppWrite,AppRead;
	si.hStdOutput = ::GetStdHandle(STD_OUTPUT_HANDLE);
	si.hStdError = ::GetStdHandle(STD_ERROR_HANDLE);
	if(m_info.RedirectStandardOutput || m_info.RedirectStandardError){
		HANDLE MyRead,MyReadTmp;
		if(!::CreatePipe(&MyReadTmp,&AppWrite,&sa,0)){SetLastError();return false;}
		if(!::DuplicateHandle(::GetCurrentProcess(),MyReadTmp,::GetCurrentProcess(),&MyRead,0,FALSE,DUPLICATE_SAME_ACCESS)){SetLastError();return false;}
		::CloseHandle(MyReadTmp);
		if(m_info.RedirectStandardOutput)si.hStdOutput = AppWrite;
		if(m_info.RedirectStandardError)si.hStdError = AppWrite;
		m_stdout = MyRead;
	}
	si.hStdInput = ::GetStdHandle(STD_INPUT_HANDLE);
	if(m_info.RedirectStandardInput){
		HANDLE MyWrite,MyWriteTmp;
		if(!::CreatePipe(&AppRead,&MyWriteTmp,&sa,0)){SetLastError();return false;}
		if(!::DuplicateHandle(::GetCurrentProcess(),MyWriteTmp,::GetCurrentProcess(),&MyWrite,0,FALSE,DUPLICATE_SAME_ACCESS)){SetLastError();return false;}
		::CloseHandle(MyWriteTmp);
		si.hStdInput = AppRead;
		m_stdin = MyWrite;
	} 
	si.dwFlags = 
		(m_info.RedirectStandardInput || m_info.RedirectStandardError || m_info.RedirectStandardInput ? STARTF_USESTDHANDLES : 0) |
		(m_info.CreateNoWindow ? 0 : STARTF_USESHOWWINDOW);
	si.wShowWindow = m_info.ShowState;
	PROCESS_INFORMATION pi;
	ablib::string workdir;
	if(m_info.WorkingDirectory != _T(""))workdir = m_info.WorkingDirectory;
	else{
        TCHAR buf[4096];
        ::GetCurrentDirectory(4095,buf);
        workdir = buf;
    }
	bool rv = (::CreateProcess(m_info.FileName != _T("") ? m_info.FileName.c_str() : NULL,const_cast<TCHAR*>(m_info.Arguments.c_str()),NULL,NULL,TRUE,
		// VistaŃR\[AvP[VNconimecCfBNgȂȂD
		// DETACHED_PROCESSw肷OKH
		// DETACHED_PROCESSł͓ĂяoꂽɃR\[oĂ̂ŁCCREATE_NO_WINDOWɂĂ݂D
		CREATE_NO_WINDOW,
		NULL,workdir.c_str(),&si,&pi) != FALSE);
	SetLastError();
	if(m_info.RedirectStandardOutput || m_info.RedirectStandardError)::CloseHandle(AppWrite);
	if(m_info.RedirectStandardInput)::CloseHandle(AppRead);
	if(!rv){
		if(m_stdout != INVALID_HANDLE_VALUE){
			::CloseHandle(m_stdout);m_stdout = INVALID_HANDLE_VALUE;
		}
		if(m_stdin != INVALID_HANDLE_VALUE){
			::CloseHandle(m_stdin);m_stdin = INVALID_HANDLE_VALUE;
		}
#ifndef NDEBUG
		DWORD d = ::GetLastError();
		DEBUGSTRING(_T("ablib::system::Process::Start : %s%s%sɂċNDsDG[R[h%d"),
			m_info.FileName.c_str(),m_info.Arguments.c_str(),m_info.WorkingDirectory.c_str(),d);
#endif
	}
	if(rv){
		m_process = pi.hProcess;
		m_thread = pi.hThread;
		m_id = pi.dwProcessId;
		m_threadid = pi.dwThreadId;
	}
	return rv;
}

void Process::Close(int exitcode){
	if(m_process != INVALID_HANDLE_VALUE){
		if(!HasExited())::TerminateProcess(m_process,exitcode);
		::CloseHandle(m_process);m_process = INVALID_HANDLE_VALUE;
		::CloseHandle(m_thread);m_thread = INVALID_HANDLE_VALUE;
	}
	if(m_stdout != INVALID_HANDLE_VALUE){
		::CloseHandle(m_stdout);m_stdout = INVALID_HANDLE_VALUE;
	}
	if(m_stdin != INVALID_HANDLE_VALUE){
		::CloseHandle(m_stdin);m_stdin = INVALID_HANDLE_VALUE;
	}
/*
	if(m_appstdout != INVALID_HANDLE_VALUE){
		::CloseHandle(m_appstdout);m_appstdout = INVALID_HANDLE_VALUE;
	}
	if(m_appstdin != INVALID_HANDLE_VALUE){
		::CloseHandle(m_appstdin);m_appstdin = INVALID_HANDLE_VALUE;
	}
*/
}

void Process::Kill(int exitcode){Close(exitcode);}


void Process::KillChilds(HANDLE handle,DWORD id,int exitcode){
	if(handle == INVALID_HANDLE_VALUE)return;
	HANDLE snapshot = ::CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS,0);
	::TerminateProcess(handle,exitcode);
	if(snapshot != INVALID_HANDLE_VALUE){
		PROCESSENTRY32 procent;
		procent.dwSize = sizeof(PROCESSENTRY32);
		BOOL b = ::Process32First(snapshot,&procent);
		while(b){
			if(procent.th32ParentProcessID == id){
				HANDLE chhandle = ::OpenProcess(PROCESS_ALL_ACCESS,TRUE,procent.th32ProcessID);
				KillChilds(chhandle,procent.th32ProcessID,exitcode);
			}
			b = ::Process32Next(snapshot,&procent);
		}
		::CloseHandle(snapshot);
	}
	return;
}

/*
DWORD Process::Read(char *buf,DWORD size){
//	if(IsExited())return 0;
	DWORD rsize;
	if(::PeekNamedPipe(m_stdout,NULL,0,NULL,&rsize,NULL) && rsize != 0){
		if(!::ReadFile(m_stdout,buf,Min(rsize,size),&rsize,NULL))return 0;
		else return rsize;
	}else return 0;
}
*/
DWORD Process::Read(char *buf,DWORD size){
	DWORD rsize;
	if(!::PeekNamedPipe(m_stdout,NULL,0,NULL,&rsize,NULL)){SetLastError();return 0;}
//	if(rsize == 0){DEBUGSTRING(_T("c0"));return 0;}
	if(rsize == 0)return 0;
	if(!::ReadFile(m_stdout,buf,Min(rsize,size),&rsize,NULL)){SetLastError();return 0;}
//	DEBUGSTRING(_T("߂l%d"),rsize);
	return rsize;
}
DWORD Process::WaitRead(char *buf,DWORD size){
	DWORD rsize;
	if(!::ReadFile(m_stdout,buf,size,&rsize,NULL)){SetLastError();return 0;}
//	DEBUGSTRING(_T("߂l%d"),rsize);
	return rsize;
}

DWORD Process::Write(const char *buf,DWORD size){
	if(HasExited())return 0;
	DWORD wsize;
	if(!::WriteFile(m_stdin,buf,size,&wsize,NULL)){SetLastError();return 0;}
	else return wsize;
}

bool Process::HasExited(){
//	DEBUGSTRING(_T("process%d"),m_process);
	if(m_process == INVALID_HANDLE_VALUE)return true;
	switch(::WaitForSingleObject(m_process,0)){
	case WAIT_ABANDONED:
	case WAIT_TIMEOUT:	return false;
	case WAIT_OBJECT_0:	return true;
	default:
		SetLastError();
		return true;
	}
}


}//namespace system
}//namespace ablib
