#ifndef __ABLIB_SYSTEM_PROCESS_H__
#define __ABLIB_SYSTEM_PROCESS_H__

#include "ablib_system.h"

namespace ablib{
namespace system{


class Process
{
public:
	Process() :
		m_thread(INVALID_HANDLE_VALUE),m_id(0),m_process(INVALID_HANDLE_VALUE),
		m_stdin(INVALID_HANDLE_VALUE),m_stdout(INVALID_HANDLE_VALUE),m_stderr(INVALID_HANDLE_VALUE),
		m_stillalive(false){}
	~Process(){if(!m_stillalive)Close();}
	struct ProcessStartInfo{
		ProcessStartInfo() : RedirectStandardOutput(false),RedirectStandardInput(false),RedirectStandardError(false),CreateNoWindow(false),ShowState(show){}
		ablib::string FileName,Arguments,WorkingDirectory;
		bool RedirectStandardOutput,RedirectStandardInput,RedirectStandardError;
		bool CreateNoWindow;
		enum showstate{show = SW_SHOW,hide = SW_HIDE};
		showstate ShowState;
	};

	void StartInfo(const ProcessStartInfo &s){m_info = s;}
	ProcessStartInfo StartInfo() const{return m_info;}
	bool Start();
	bool Start(const ProcessStartInfo &s){StartInfo(s);return Start();}
	void Kill(int exitcode = -1);
	bool WaitForExit(int time = INFINITE){return ::WaitForSingleObject(m_process,time) == WAIT_OBJECT_0;}
	bool WaitForInputIdle(int time = INFINITE){return ::WaitForInputIdle(m_process,time) == 0;}
	void SetStillAlive(bool b){m_stillalive = b;}

	enum wait_msg{timeout,message,process_exit};
	wait_msg WaitForMsg(int time = INFINITE){
		switch(::MsgWaitForMultipleObjects(1,&m_thread,FALSE,time,QS_ALLEVENTS)){
		case WAIT_OBJECT_0:		return process_exit;
		case WAIT_OBJECT_0 + 1:	return message;
		default:				return timeout;
		}
	}

	DWORD Read(char *buf,DWORD size);
	DWORD WaitRead(char *buf,DWORD size);
	DWORD Write(const char *buf,DWORD size);
	DWORD Write(const char *buf){return Write(buf,::lstrlenA(buf));}
	DWORD ExitCode(){DWORD rv;::GetExitCodeProcess(m_process,&rv);return rv;}
	
	bool HasExited();
	
	HANDLE GetThread() const{return m_thread;}
	HANDLE GetProcess() const{return m_process;}
	DWORD Id() const{return m_id;}
	DWORD GetThreadID() const{return m_threadid;}
	
	void KillChilds(int exitcode = -1){KillChilds(m_process,m_id,exitcode);Kill(exitcode);}
	
private:
	void KillChilds(HANDLE processhandle,DWORD processid,int exitcode);

	void Close(int exitcode = -1);
	ProcessStartInfo m_info;
	HANDLE m_thread,m_process,m_stdin,m_stdout,m_stderr;
	DWORD m_id,m_threadid;
	bool m_stillalive;
};

}//namespace system
}//namespace ablib

#endif //__ABLIB_SYSTEM_PROCESS_H__

