#ifndef __ABLIB_SYSTEM_TIME_H__
#define __ABLIB_SYSTEM_TIME_H__

#include "ablib_system.h"

namespace ablib{
namespace system{

class Time
{
public:
	Time(const SYSTEMTIME &st) : m_st(st),m_valid(true){}
	Time(const FILETIME &ft) : m_valid(true){::FileTimeToSystemTime(&ft,&m_st);}
	Time(int year,int month,int day,int hour = 0,int minute = 0,int second = 0,int millisecond = 0);
	Time() : m_valid(false){}
	~Time(){}


	SYSTEMTIME SystemTime(){return m_st;}

	bool IsValid() const{return m_valid;}
	void SetValid(bool b){m_valid = b;}
	ablib::string ToStr();
	ablib::string RFC1123String();

	bool operator<(const Time &t) const;
	bool operator>(const Time &t) const{return !(operator<(t) || operator==(t));}
	bool operator==(const Time &t) const;
	
	Time &operator+=(const Time &obj);

	DWORD year() const{return m_st.wYear;}
	DWORD month() const{return m_st.wMonth;}
	DWORD day() const{return m_st.wDay;}
	DWORD hour() const{return m_st.wHour;}
	DWORD minute() const{return m_st.wMinute;}
	DWORD second() const{return m_st.wSecond;}

	void GetLocalTime(){::GetLocalTime(&m_st);}
	void GetSystemTime(){::GetSystemTime(&m_st);}

	bool CheckValid();
	void SetYoubi(){m_st.wDayOfWeek = GetYoubi(m_st.wYear,m_st.wMonth,m_st.wDay);}

	static bool IsUruu(int year);
	static WORD GetYoubi(int year,int month,int day);

private:
	SYSTEMTIME m_st;
	bool m_valid;
	friend class File;
};

inline const Time operator+(const Time &lhs,const Time &rhs){return Time(lhs) += rhs;}

}//namespace system
}//namespace ablib

#endif //__ABLIB_SYSTEM_TIME_H__

