#include "Translation.h"
#include <map>
#include <ablib/lexical_cast.h>
#include <system/FindFile.h>

namespace ablib{
namespace system{
namespace Translation{

namespace impl{

std::map<ablib::string,int> g_str2id;
HMODULE g_resmod = NULL;
LANGID g_langid = 0,g_sourceid = 0;
ablib::string g_dll;
ablib::string tmp_string;

BOOL CALLBACK EnumResNameProc(HMODULE hModule,LPCTSTR lpszType,LPTSTR lpszName,LONG_PTR lParam){
	if(lpszType == RT_STRING){
		if(HIWORD(lpszName) == 0){
			ablib::string str;
			HRSRC hRc = ::FindResourceEx(hModule,RT_STRING,lpszName,MAKELANGID(LANG_JAPANESE,SUBLANG_JAPANESE_JAPAN));
			if(hRc){
				TCHAR *mem = reinterpret_cast<TCHAR*>(::LoadResource(hModule,hRc));
				if(mem){
					TCHAR *memcur = mem;
					for(int i = 0 ; i < 16 ; ++i){
						if(*memcur){
							int len = *memcur;
							++memcur;
							str.assign(memcur,len);
							g_str2id[str] = (LOWORD(lpszName) - 1)*16 + i;
							memcur += len;
						}else ++memcur;
					}
				}
			}
		}
	}
	return TRUE;
}

// ^`
typedef struct LANGINFO_DEF {
	int		Count;
	LANGID	LangID;
} LANGINFO;
typedef LANGINFO *PLANGINFO;

BOOL CALLBACK EnumLangProc(HANDLE, LPCTSTR, LPCTSTR,
						   WORD wIDLanguage, LONG_PTR lParam)
{
    PLANGINFO LangInfo;

    LangInfo = (PLANGINFO) lParam;
    LangInfo->Count++;
    LangInfo->LangID  = wIDLanguage;

    return (TRUE);        // 񋓂p܂
}


LANGID GetNTDLLNativeLangID()
{

    LANGINFO LangInfo;
	LPCTSTR Type = (LPCTSTR) ((LPVOID)((WORD)16));
    LPCTSTR Name = (LPCTSTR) 1;

    ZeroMemory(&LangInfo,sizeof(LangInfo));
    
    // ntdll  HModule 擾܂
    HMODULE hMod = GetModuleHandle(_T("ntdll.dll"));
    if (hMod==NULL) {
        return(0);
    }

    BOOL result = EnumResourceLanguages(hMod, Type, Name, (ENUMRESLANGPROC)EnumLangProc, (LONG_PTR) &LangInfo);
    
    if (!result || (LangInfo.Count > 2) || (LangInfo.Count < 1) ) {
        return (0);
    }
    
    return (LangInfo.LangID);
}


BOOL IsHongKongVersion()
{
    HMODULE hMod;
    BOOL bRet=FALSE;
	typedef BOOL (WINAPI *IMMRELEASECONTEXT)(HWND,HIMC);
    IMMRELEASECONTEXT pImmReleaseContext;

    hMod = LoadLibrary(_T("imm32.dll"));
    if (hMod) {
        pImmReleaseContext = (IMMRELEASECONTEXT)GetProcAddress(hMod,"ImmReleaseContext");
        if (pImmReleaseContext) {
            bRet = pImmReleaseContext(NULL,NULL);
        }
        FreeLibrary(hMod);
    }
    return (bRet);
}


LANGID DetectLanguage() {
#define MAX_KEY_BUFFER	80
	OSVERSIONINFO		VersionInfo;
	LANGID				uiLangID = 0;
	HKEY				hKey;
	DWORD				Type, BuffLen = MAX_KEY_BUFFER;
	TCHAR				LangKeyValue[MAX_KEY_BUFFER];

	VersionInfo.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	if(!GetVersionEx(&VersionInfo))return 0;

	switch( VersionInfo.dwPlatformId ) {
		// Windows NTAWindows 2000A܂͂ȏ̏ꍇ
		case VER_PLATFORM_WIN32_NT:
			if( VersionInfo.dwMajorVersion >= 5)   // Windows 2000 ܂͂ȏ
				uiLangID = GetUserDefaultUILanguage();
			else {   // ntdll.dll ̌ NT4 `FbNp
				uiLangID = GetNTDLLNativeLangID();   
				if (uiLangID == 1033) {		// NT 4  Honkong SAR o[W̓ʂȏ
					if (IsHongKongVersion()) {
						uiLangID = 3076;
					}
				}
			}
			break;
		// Windows 98 ܂ Windows Me ̏ꍇ
		case VER_PLATFORM_WIN32_WINDOWS:
			// UI ̃WXg L[J܂
			if( RegOpenKeyEx(HKEY_CURRENT_USER,_T("Default\\Control Panel\\Desktop\\ResourceLocale"), 0,
				KEY_QUERY_VALUE, &hKey) == ERROR_SUCCESS ) {
				// ̃L[̌^擾܂
				if( RegQueryValueEx(hKey, NULL, NULL, &Type, NULL, NULL) == ERROR_SUCCESS 
					&& Type == REG_SZ ) {
					// L[̒lǂݎ܂
					if( RegQueryValueEx(hKey, NULL, NULL, &Type, (LPBYTE)LangKeyValue, &BuffLen) 
						== ERROR_SUCCESS ) {
						uiLangID = (LANGID)_ttoi(LangKeyValue);
					}
				}
				RegCloseKey(hKey);
			}				
			break;
	}

    if (uiLangID == 0) {
        uiLangID = GetUserDefaultLangID();
    }
    //  ID Ԃ܂B
    return (uiLangID);
}

ablib::string GetLangDir(){
	TCHAR buf[4096];
	::GetModuleFileName(NULL,buf,4095);
	ablib::string dir = buf;
	ablib::string::size_type r = dir.rfind(_T("\\"));
	if(r != ablib::string::npos)dir.erase(r,ablib::string::npos);
	return dir + _T("\\lang\\");
}

// lang\(num).dll
HMODULE LoadLangDLL(LANGID lang){
	if(g_dll != _T(""))::FreeLibrary(g_resmod);
	g_dll = GetLangDir() + ablib::lexical_cast<ablib::string>(lang) + _T(".dll");
	HMODULE rv = ::LoadLibraryEx(g_dll.c_str(),NULL,LOAD_LIBRARY_AS_DATAFILE);
	if(!rv)g_dll = _T("");
	return rv;
}

}//namespace impl

HINSTANCE InitTranslation(LANGID sourceid){
	impl::g_sourceid = sourceid;
	::EnumResourceNames(NULL,RT_STRING,impl::EnumResNameProc,0);
	LANGID langid = impl::DetectLanguage();
//	langid = 1033;
	HMODULE mod = impl::LoadLangDLL(langid);
	if(mod){
		impl::g_resmod = mod;
		impl::g_langid = langid;
	}else{
		mod = ::GetModuleHandle(NULL);
		impl::g_langid = impl::g_sourceid;
	}
//	::SetThreadLocale(impl::g_langid);
	return impl::g_resmod;
}

ablib::string tmp_string;
const TCHAR *GetTranslation(const TCHAR *str){
    TCHAR buf[4096];
	if(::LoadString(impl::g_resmod,impl::g_str2id[str],buf,4095)){
        tmp_string = buf;
        return tmp_string.c_str();
	}else return str;
}

const TCHAR *GetTranslation(UINT id){
	TCHAR buf[4096];
	if(::LoadString(impl::g_resmod,id,buf,4095)){
        tmp_string = buf;
        return tmp_string.c_str();
	}else if(::LoadString(::GetModuleHandle(NULL),id,buf,4095)){
        tmp_string = buf;
        return tmp_string.c_str();
	}else return _T("translation is not found");
}

HINSTANCE GetResourceHandle(){return impl::g_resmod;}
LANGID GetLanguage(){return impl::g_langid;}
HINSTANCE SetLanguage(LANGID langid){
	HMODULE mod = impl::LoadLangDLL(langid);
	if(mod){
		impl::g_resmod = mod;
		impl::g_langid = langid;
	}else{
		impl::g_resmod = ::GetModuleHandle(NULL);
		impl::g_langid = impl::g_sourceid;
	}
//	::SetThreadLocale(impl::g_langid);
	return impl::g_resmod;
}
std::vector<LangInfo> GetAvailableLanguages(){
	std::vector<LangInfo> rv;
	TCHAR buf[256];
	if(::GetLocaleInfo(impl::g_sourceid,LOCALE_SNATIVELANGNAME,buf,255)){
		rv.push_back(LangInfo(buf,impl::g_sourceid));
	}

	ablib::system::FindFile ff;
	if(ff.Begin(impl::GetLangDir() + _T("*.dll"))){
		WIN32_FIND_DATA wfd;
		while(ff.Next(&wfd)){
			LANGID id = ::_tstoi(wfd.cFileName);
			if(id != 0 && ::GetLocaleInfo(id,LOCALE_SNATIVELANGNAME,buf,255)){
				rv.push_back(LangInfo(buf,id));
			}
		}
	}
	return rv;
}

}//namespace Translation
}//namespace system
}//namespace ablib

