#include "ablib_window.h"
#include "BaseWnd.h"


namespace ablib{
namespace window{

BaseWnd::~BaseWnd(){
	if(!::FindWindow(m_wclassname.c_str(),NULL))::UnregisterClass(m_wclassname.c_str(),m_appint);
	delete m_window;
}

void BaseWnd::AddSubClass(SubClassDlg *sub,int id){
	m_subclass.push_back(SubClassData(sub,id));
	if(m_window){
		unsigned int i = m_subclass.size() - 1;
		m_subclass[i].hwnd = ::GetDlgItem(m_window->GetHWND(),id);
		m_subclass[i].subclass->SubClass(m_subclass[i].hwnd);
	}
}
void BaseWnd::AddSubClass(SubClassDlg *sub,HWND hwnd){
	m_subclass.push_back(SubClassData(sub,hwnd));
	if(m_window){
		unsigned int i = m_subclass.size() - 1;
		m_subclass[i].subclass->SubClass(m_subclass[i].hwnd);
	}
}

Window *BaseWnd::Create(Wnd *parent,HCURSOR cursor){
	if(m_window)return NULL;
	WNDCLASSEX wc;
	::ZeroMemory(&wc,sizeof(WNDCLASSEX));
	if(::GetClassInfoEx(m_appint,m_wclassname.c_str(),&wc) == 0){
		wc.style = CS_DBLCLKS;
		wc.cbSize = sizeof(WNDCLASSEX);
		wc.hInstance = m_appint;
		wc.lpfnWndProc = sWndProc;
		wc.lpszClassName = m_wclassname.c_str();
		wc.hCursor = cursor;
		wc.hbrBackground = reinterpret_cast<HBRUSH>(::GetStockObject(WHITE_BRUSH));

		if(!::RegisterClassEx(&wc))return NULL;
	}
	HWND hwnd = ::CreateWindow(m_wclassname.c_str(),_T(""),m_style ,CW_USEDEFAULT,0,CW_USEDEFAULT,0,(parent ? parent->GetHWND() : NULL),NULL,m_appint,this);
	if(hwnd == NULL){
		DEBUGSTRING(_T("ablib::window::BaseWnd::Create : EBhEsDG[R[hF%lu"),::GetLastError());
		return NULL;
	}

	DEBUGSTRING(_T("ablib::window::BaseWnd::Create : EBhEDnh%lu NX%s"),hwnd,m_wclassname.c_str());

//	m_window->ShowWindow(Window::sw_show);
	return m_window;
}

LRESULT CALLBACK BaseWnd::sWndProc(HWND hwnd,UINT uMsg,WPARAM wParam,LPARAM lParam){
	BaseWnd *target = static_cast<BaseWnd*>(::GetProp(hwnd,_T("ABENORI_WINDOW_PROP")));
	if(!target){
		if(uMsg == WM_CREATE){
			target = static_cast<BaseWnd*>(reinterpret_cast<LPCREATESTRUCT>(lParam)->lpCreateParams);
			target->Atach(hwnd);
			target->m_window = new Window(hwnd);
		}
	}
	if(target){
		Window *me = target->GetWindow();
		if(!me)return ::DefWindowProc(hwnd,uMsg,wParam,lParam);
		
		{
			LRESULT rv = TRUE;
			if(target->PreTranslateMessage(me,uMsg,wParam,lParam,rv))return rv;
		}
		
		switch(uMsg){
		case WM_CREATE:
		{
			intptr_t rv = target->OnCreate(me) ? 0 : ::DefWindowProc(hwnd,uMsg,wParam,lParam);
			for(intptr_t i = 0 ; i < target->m_subclass.size() ; ++i){
				int id = target->m_subclass[(int)i].id;
				target->m_subclass[(int)i].subclass->SubClass(::GetDlgItem(hwnd,id));
			}
			return rv;
		}
		case WM_CLOSE:	return target->OnClose(me) ? 0 : ::DefWindowProc(hwnd,uMsg,wParam,lParam);
		case WM_DESTROY:
			if(target->OnDestroy(me)){
				target->Detach(hwnd);
				delete target->m_window;
				target->m_window = NULL;
				for(int i = 0 ; i < target->m_subclass.size() ; ++i){
					target->m_subclass[i].subclass->UnSubClass();
				}
				if(target->m_stopmsgloop)::PostQuitMessage(0);
				return 0;
			}else return ::DefWindowProc(hwnd,uMsg,wParam,lParam);
		case WM_PAINT:	return target->OnPaint(me) ? 0 : ::DefWindowProc(hwnd,uMsg,wParam,lParam);;
		case WM_SIZE:	return target->OnSize(me,wParam,LOWORD(lParam),HIWORD(lParam)) ? 0 : ::DefWindowProc(hwnd,uMsg,wParam,lParam);;
		case WM_SETFOCUS:
		{
			Window w(reinterpret_cast<HWND>(wParam));
			return target->OnSetFocus(me,&w) ? 0 : ::DefWindowProc(hwnd,uMsg,wParam,lParam);
		}
		case WM_KILLFOCUS:
		{
			Window w(reinterpret_cast<HWND>(wParam));
			return target->OnKillFocus(me,&w) ? 0 : ::DefWindowProc(hwnd,uMsg,wParam,lParam);
		}
		case WM_MOUSEMOVE:
		{
			Mouse m(hwnd,Point(static_cast<int>(static_cast<short>(LOWORD(lParam))),static_cast<int>(static_cast<short>(HIWORD(lParam)))));
			return target->OnMouseMove(me,&m,static_cast<int>(wParam)) ? 0 : ::DefWindowProc(hwnd,uMsg,wParam,lParam);
		}
		case WM_LBUTTONDOWN:
		{
			Mouse m(hwnd,Point(static_cast<int>(static_cast<short>(LOWORD(lParam))),static_cast<int>(static_cast<short>(HIWORD(lParam)))));
			return target->OnLButtonDown(me,&m,false) ? 0 : ::DefWindowProc(hwnd,uMsg,wParam,lParam);
		}
		case WM_LBUTTONDBLCLK:
		{
			Mouse m(hwnd,Point(static_cast<int>(static_cast<short>(LOWORD(lParam))),static_cast<int>(static_cast<short>(HIWORD(lParam)))));
			return target->OnLButtonDown(me,&m,true) ? 0 : ::DefWindowProc(hwnd,uMsg,wParam,lParam);
		}
		case WM_LBUTTONUP:
		{
			Mouse m(hwnd,Point(static_cast<int>(static_cast<short>(LOWORD(lParam))),static_cast<int>(static_cast<short>(HIWORD(lParam)))));
			return target->OnLButtonUp(me,&m) ? 0 : ::DefWindowProc(hwnd,uMsg,wParam,lParam);
		}
		case WM_RBUTTONDOWN:
		{		
			Mouse m(hwnd,Point(static_cast<int>(static_cast<short>(LOWORD(lParam))),static_cast<int>(static_cast<short>(HIWORD(lParam)))));
			return target->OnRButtonDown(me,&m,false) ? 0 : ::DefWindowProc(hwnd,uMsg,wParam,lParam);
		}
		case WM_RBUTTONDBLCLK:
		{
			Mouse m(hwnd,Point(static_cast<int>(static_cast<short>(LOWORD(lParam))),static_cast<int>(static_cast<short>(HIWORD(lParam)))));
			return target->OnRButtonDown(me,&m,true) ? 0 : ::DefWindowProc(hwnd,uMsg,wParam,lParam);
		}
		case WM_RBUTTONUP:
		{
			Mouse m(hwnd,Point(static_cast<int>(static_cast<short>(LOWORD(lParam))),static_cast<int>(static_cast<short>(HIWORD(lParam)))));
			return target->OnRButtonUp(me,&m) ? 0 : ::DefWindowProc(hwnd,uMsg,wParam,lParam);
		}
		case WM_MBUTTONDOWN:
		{
			Mouse m(hwnd,Point(static_cast<int>(static_cast<short>(LOWORD(lParam))),static_cast<int>(static_cast<short>(HIWORD(lParam)))));
			return target->OnMButtonDown(me,&m,false) ? 0 : ::DefWindowProc(hwnd,uMsg,wParam,lParam);
		}
		case WM_MBUTTONDBLCLK:
		{
			Mouse m(hwnd,Point(static_cast<int>(static_cast<short>(LOWORD(lParam))),static_cast<int>(static_cast<short>(HIWORD(lParam)))));
			return target->OnMButtonDown(me,&m,true) ? 0 : ::DefWindowProc(hwnd,uMsg,wParam,lParam);
		}
		case WM_MBUTTONUP:
		{
			Mouse m(hwnd,Point(static_cast<int>(static_cast<short>(LOWORD(lParam))),static_cast<int>(static_cast<short>(HIWORD(lParam)))));
			return target->OnMButtonUp(me,&m) ? 0 : ::DefWindowProc(hwnd,uMsg,wParam,lParam);
		}
		case WM_MOUSEWHEEL:
		{
			Mouse m(hwnd,Point(static_cast<int>(static_cast<short>(LOWORD(lParam))),static_cast<int>(static_cast<short>(HIWORD(lParam)))));
			return target->OnMouseWheel(me,&m,static_cast<short>(HIWORD(wParam)) / WHEEL_DELTA,static_cast<int>(LOWORD(wParam))) ? 0 : ::DefWindowProc(hwnd,uMsg,wParam,lParam);
		}
		case WM_KEYDOWN:		return target->OnKey(me,true,static_cast<UINT>(wParam),static_cast<int>(LOWORD(lParam))) ? 0 : ::DefWindowProc(hwnd,uMsg,wParam,lParam);
		case WM_KEYUP:			return target->OnKey(me,false,static_cast<UINT>(wParam),static_cast<int>(LOWORD(lParam))) ? 0 : ::DefWindowProc(hwnd,uMsg,wParam,lParam);
		case WM_CHAR:			return target->OnChar(me,static_cast<TCHAR>(wParam),static_cast<int>(LOWORD(lParam))) ? 0 : ::DefWindowProc(hwnd,uMsg,wParam,lParam);
		case WM_IME_COMPOSITION:
			if(lParam & GCS_RESULTSTR)return target->OnIME(me) ? 0 : ::DefWindowProc(hwnd,uMsg,wParam,lParam);
			else return ::DefWindowProc(hwnd,uMsg,wParam,lParam);
		case WM_IME_CHAR:		return 0;
		case WM_COMMAND:		return target->OnCommand(me,static_cast<int>(LOWORD(wParam)),static_cast<int>(HIWORD(wParam))) ? 0 : ::DefWindowProc(hwnd,uMsg,wParam,lParam);
		case WM_NOTIFY:
		{
			LRESULT rv = TRUE;
			if(target->OnNotify(me,reinterpret_cast<NMHDR*>(lParam),rv))return rv;
			else return ::DefWindowProc(hwnd,uMsg,wParam,lParam);
		}
		case WM_HSCROLL:		return target->OnHScroll(me,target->CalcScroll(SB_HORZ,(int)(short)LOWORD(wParam))) ? 0 : ::DefWindowProc(hwnd,uMsg,wParam,lParam);
		case WM_VSCROLL:		return target->OnVScroll(me,target->CalcScroll(SB_VERT,(int)(short)LOWORD(wParam))) ? 0 : ::DefWindowProc(hwnd,uMsg,wParam,lParam);
		case WM_DROPFILES:
		{
			std::vector<ablib::string> files;
			int n = ::DragQueryFile(reinterpret_cast<HDROP>(wParam),-1,NULL,0);
			for(int i = 0 ; i < n ; ++i){
				int size = ::DragQueryFile(reinterpret_cast<HDROP>(wParam),i,NULL,0);
				TCHAR *buf = new TCHAR[size + 1];
				::DragQueryFile(reinterpret_cast<HDROP>(wParam),i,buf,size + 1);
				files.push_back(buf);
				delete[] buf;
			}
			::DragFinish(reinterpret_cast<HDROP>(wParam));
			target->OnDropFiles(files);
			return 0;
		}
		default:				return ::DefWindowProc(hwnd,uMsg,wParam,lParam);
		}
	}else return ::DefWindowProc(hwnd,uMsg,wParam,lParam);
}

int BaseWnd::CalcScroll(LONG type,UINT code){
	SCROLLINFO si;
	::ZeroMemory(&si,sizeof(SCROLLINFO));
	si.cbSize = sizeof(SCROLLINFO);
	si.fMask = SIF_ALL;
	::GetScrollInfo(m_window->GetHWND(),type,&si);
	switch(code){
	case SB_LINEUP:		return -1;
	case SB_LINEDOWN:	return 1;
	case SB_PAGEUP:		return -static_cast<int>(si.nPage);
	case SB_PAGEDOWN:	return si.nPage;
	case SB_TOP:		return si.nMin - si.nPos;
	case SB_BOTTOM:		return si.nMax - si.nPos;
	case SB_THUMBTRACK:	return si.nTrackPos - si.nPos;
	default:			return 0;
	}
}

}//namespace windows
}//namespace arch
