#include "ablib_window.h"
#include "window.h"

namespace ablib{
namespace window{

Point Caret::GetPos(){
	POINT pt;
	::GetCaretPos(&pt);
	::ClientToScreen(m_hwnd,&pt);
	return Point(pt.x,pt.y);
}

void IME::SetPos(Point pt){
	HIMC ime;
	COMPOSITIONFORM cf;
	cf.dwStyle = CFS_POINT;
	cf.ptCurrentPos.x = (int)pt.x;
	cf.ptCurrentPos.y = (int)pt.y;
	ime = ::ImmGetContext(m_hwnd);
	::ImmSetCompositionWindow(ime,&cf);
	::ImmReleaseContext(m_hwnd,ime);
	return;
}

void IME::SetFont(FontData f){
	HIMC ime = ::ImmGetContext(m_hwnd);
	LOGFONT lf;

	lf.lfHeight = 2 * ((LONG)f.height / 2);
	lf.lfWidth = (LONG)f.avewidth;
	lf.lfEscapement = 0;
	lf.lfOrientation = 0;
	lf.lfWeight = (LONG)f.weight;
	lf.lfItalic = FALSE;
	lf.lfUnderline = FALSE;
	lf.lfStrikeOut = FALSE;
//	lf.lfCharSet = DEFAULT_CHARSET;
	lf.lfCharSet = SHIFTJIS_CHARSET;
	lf.lfOutPrecision = OUT_DEFAULT_PRECIS;
	lf.lfClipPrecision = CLIP_DEFAULT_PRECIS;
	lf.lfQuality = PROOF_QUALITY;
	lf.lfPitchAndFamily = FIXED_PITCH | FF_MODERN;
	::lstrcpyn(lf.lfFaceName,f.fontname.c_str(),32);

	::ImmSetCompositionFont(ime,&lf);
	::ImmReleaseContext(m_hwnd,ime);
	return;
}

bool IME::GetText(string &str){
	HIMC ime = ::ImmGetContext(m_hwnd);
	LONG size = ::ImmGetCompositionString(ime,GCS_RESULTSTR,NULL,0);
	if(size == IMM_ERROR_NODATA || size == IMM_ERROR_GENERAL){
		::ImmReleaseContext(m_hwnd,ime);
		return false;
	}
	TCHAR *buf = new TCHAR[size + 1];
	size = ::ImmGetCompositionString(ime,GCS_RESULTSTR,buf,size);
	if(size == IMM_ERROR_NODATA || size == IMM_ERROR_GENERAL){
		::ImmReleaseContext(m_hwnd,ime);
		delete[] buf;
		return false;
	}
	buf[size] = '\0';
	str = buf;
	delete[] buf;
	::ImmReleaseContext(m_hwnd,ime);
	return true;
}

bool IME::IsOpen(){
	HIMC ime = ::ImmGetContext(m_hwnd);
	int rv = ::ImmGetOpenStatus(ime);
	::ImmReleaseContext(m_hwnd,ime);
	return rv ? true : false;
}

void IME::SetOpen(bool b){
	HIMC ime = ::ImmGetContext(m_hwnd);
	::ImmSetOpenStatus(ime,b ? TRUE : FALSE);
	::ImmReleaseContext(m_hwnd,ime);
}

bool ClipBoard::GetText(string &str){
	if(!::IsClipboardFormatAvailable(CF_TEXT))return false;
	HGLOBAL mem = static_cast<HGLOBAL>(::GetClipboardData(CF_TEXT));
	if(!mem)return false;
	TCHAR *p = static_cast<TCHAR*>(::GlobalLock(mem));
	str = p;
	::GlobalUnlock(mem);
	return true;
}

bool ClipBoard::SetText(const TCHAR *str,size_t len){
	if(!::EmptyClipboard())return false;
	HGLOBAL mem = ::GlobalAlloc(GMEM_MOVEABLE,(len + 1)*sizeof(TCHAR));
	if(!mem)return false;
	TCHAR *p = static_cast<TCHAR*>(::GlobalLock(mem));
	if(!p){
		::GlobalFree(mem);
		return false;
	}
	::lstrcpy(p,str);
	::GlobalUnlock(mem);
#ifdef UNICODE
	if(!::SetClipboardData(CF_UNICODETEXT,mem)){
#else
	if(!::SetClipboardData(CF_TEXT,mem)){
#endif
		::GlobalFree(mem);
		return false;
	}
	return true;
}

bool ClipBoard::SetFile(const std::vector<ablib::string> &files){
	if(!::EmptyClipboard())return false;
	size_t len = 0;
	for(const ablib::string &s : files) { len += (s.length() + 1); }
	HGLOBAL mem = ::GlobalAlloc(GHND,sizeof(DROPFILES) + sizeof(TCHAR) * len + 10);
	if(!mem)return false;
	LPDROPFILES dfile = static_cast<LPDROPFILES>(::GlobalLock(mem));
	dfile->pFiles = sizeof(DROPFILES);
	dfile->pt.x = 0;dfile->pt.y = 0;
	dfile->fNC = FALSE;
#ifdef UNICODE
	dfile->fWide = TRUE;
#else
	dfile->fWide = FALSE;
#endif
	TCHAR *fnames = reinterpret_cast<TCHAR*>(dfile + 1);
	for(const ablib::string &s : files){::lstrcpy(fnames,s.c_str());fnames += (s.length() + 1);DEBUGSTRING(_T("%d"),s.length());}
	*fnames = '\0';
#ifndef UNICODE
	++fnames;*fnames = '\0';
#endif
	::GlobalUnlock(mem);
	if(!::SetClipboardData(CF_HDROP,mem)){
		::GlobalFree(mem);
		return false;
	}
	return true;
}

void ScrollBarimpl<true>::SetRange(intptr_t from, intptr_t to){
	m_si.nMin = (int)from;m_si.nMax = (int)to;
	::SetScrollInfo(m_hwnd,SB_VERT,&m_si,TRUE);
	::GetScrollInfo(m_hwnd,SB_VERT,&m_si);
}
void ScrollBarimpl<false>::SetRange(intptr_t from, intptr_t to){
	m_si.nMin = (int)from;m_si.nMax = (int)to;
	::SetScrollInfo(m_hwnd,SB_HORZ,&m_si,TRUE);
	::GetScrollInfo(m_hwnd,SB_HORZ,&m_si);
}

void ScrollBarimpl<true>::SetPage(intptr_t page){
	m_si.nPage = (UINT)page;
	::SetScrollInfo(m_hwnd,SB_VERT,&m_si,TRUE);
	::GetScrollInfo(m_hwnd,SB_VERT,&m_si);
}
void ScrollBarimpl<false>::SetPage(intptr_t page){
	m_si.nPage = (UINT)page;
	::SetScrollInfo(m_hwnd,SB_HORZ,&m_si,TRUE);
	::GetScrollInfo(m_hwnd,SB_HORZ,&m_si);
}

void ScrollBarimpl<true>::SetPos(intptr_t pos){
	m_si.nPos = (int)pos;
	::SetScrollInfo(m_hwnd,SB_VERT,&m_si,TRUE);
	::GetScrollInfo(m_hwnd,SB_VERT,&m_si);
}
void ScrollBarimpl<false>::SetPos(intptr_t pos){
	m_si.nPos = (int)pos;
	::SetScrollInfo(m_hwnd,SB_HORZ,&m_si,TRUE);
	::GetScrollInfo(m_hwnd,SB_HORZ,&m_si);
}

void ScrollBarimpl<true>::WindowScroll(intptr_t scroll){
	::ScrollWindowEx(m_hwnd,0,(int)scroll,NULL,NULL,NULL,NULL,SW_INVALIDATE);
}
void ScrollBarimpl<false>::WindowScroll(intptr_t scroll){
	::ScrollWindowEx(m_hwnd,(int)scroll,0,NULL,NULL,NULL,NULL,SW_INVALIDATE);
}

Menu::MenuItem::~MenuItem(){
	for(int i = 0 ; i < m_data.size() ; ++i)delete m_data[i].menu;
}

bool Menu::MenuItem::AddSubMenu(MenuItem *mi,int parent){
	for(int i = 0 ; i < m_data.size() ; ++i){
		if(m_data[i].menu){
			if(m_data[i].menu->AddSubMenu(mi,parent))return true;
		}else{
			if(parent == m_data[i].id){
				m_data[i].menu = mi;
				return true;
			}
		}
	}
	return false;
}

bool Menu::MenuItem::DeleteMenu(int id){
	for(int i = 0 ; i < m_data.size() ; ++i){
		if(m_data[i].id == id){
			delete m_data[i].menu;
			m_data.erase(i);
		}else{
			if(m_data[i].menu){
				if(m_data[i].menu->DeleteMenu(id))return true;
			}
		}
	}
	return false;
}

bool Menu::MenuItem::IsExistID(int id){
	if(id == 0)return false;
	for(int i = 0 ; i < m_data.size() ; ++i){
		if(m_data[i].menu){
			if(m_data[i].menu->IsExistID(id))return true;
		}
		if(m_data[i].id == id)return true;
	}
	return false;
}

Menu::MenuItem *Menu::MenuItem::GetMenuItemFromID(int id){
	MenuItem *rv;
	for(int i = 0 ; i < m_data.size() ; ++i){
		if(m_data[i].id == id)return m_data[i].menu;
		if(m_data[i].menu){
			rv = m_data[i].menu->GetMenuItemFromID(id);
			if(rv)return rv;
		}
	}
	return NULL;
}

void Menu::MenuItem::CreateMenu(HMENU hmenu,int defmenuid){
	MENUITEMINFO mii;
	::ZeroMemory(&mii,sizeof(MENUITEMINFO));
	mii.cbSize = sizeof(MENUITEMINFO);
	for(int i = 0 ; i < m_data.size() ; ++i){
		if(m_data[i].menu){
			mii.fMask = MIIM_TYPE | MIIM_SUBMENU | MIIM_ID;
			mii.hSubMenu = ::CreateMenu();
			m_data[i].menu->CreateMenu(mii.hSubMenu,defmenuid);
		}else{
			mii.fMask = MIIM_TYPE | MIIM_ID;
			mii.hSubMenu = NULL;
		}
		if(m_data[i].id != 0)mii.fType = MFT_STRING;
		else mii.fType = MFT_SEPARATOR;
		mii.wID = m_data[i].id;
		mii.dwTypeData = const_cast<TCHAR*>(m_data[i].str.c_str());
		mii.cch = (int)m_data[i].str.length();

//		DEBUGSTRING(_T("ablib::window::arch::Window::Windowmenu::MenuItem::CreateMenu : %sID%dō쐬D"),m_data[i].str.c_str(),m_data[i].id);
		
		::InsertMenuItem(hmenu,i,TRUE,&mii);
		if(m_data[i].id == defmenuid)::SetMenuDefaultItem(hmenu,i,TRUE);
	}
	return;
}

void Menu::MenuItem::Clear(){
	for(int i = 0 ; i < m_data.size() ; ++i){
		if(m_data[i].menu){
			delete m_data[i].menu;
			m_data[i].menu = NULL;
		}
	}
	m_data.clear();
}

bool Menu::SetDefaultMenu(int id){
	if(id == 0){m_defmenuid = 0;return true;}
	if(!m_menuitem.IsExistID(id))return false;
	m_defmenuid = id;
	return true;
}

bool Menu::AddMenu(int parent,int id,const TCHAR *str){
//	if(m_menuitem.IsExistID(id) && ::lstrcmp(str,_T("")) != 0)return false;
	if(!m_menuitem.IsExistID(parent))return false;
	MenuItem *mi = m_menuitem.GetMenuItemFromID(parent);
	if(mi){
		mi->AddMenu(id,str);
		return true;
	}else{
		mi = new MenuItem();
		mi->AddMenu(id,str);
		if(!m_menuitem.AddSubMenu(mi,parent)){
			delete mi;
			return false;
		}else return true;
	}
}

bool Menu::AddRootMenu(int id,const TCHAR *str){
	if(id != 0 && m_menuitem.IsExistID(id))return false;
	m_menuitem.AddMenu(id,str);
	return true;
}

bool Menu::CreateWindowMenu(){
	HMENU old = ::GetMenu(m_hwnd);
	if(old)Destroy(old);
	HMENU New = ::CreateMenu();
	m_menuitem.CreateMenu(New,m_defmenuid);

	if(::SetMenu(m_hwnd,New)){
		::DrawMenuBar(m_hwnd);
		return true;
	}else return false;
}

void Menu::Destroy(HMENU menu){
	MENUITEMINFO mii;
	::ZeroMemory(&mii,sizeof(MENUITEMINFO));
	mii.cbSize = sizeof(MENUITEMINFO);
	mii.fMask = MIIM_SUBMENU;
	int num = ::GetMenuItemCount(menu);
	for(int i = 0 ; i < num ; ++i){
		::GetMenuItemInfo(menu,i,TRUE,&mii);
		if(mii.hSubMenu)Destroy(mii.hSubMenu);
		mii.hSubMenu = NULL;
	}
	::DestroyMenu(menu);
}

int Menu::CreatePopupMenu(Point pt,posalign pos){
	HMENU menu = ::CreatePopupMenu();
	m_menuitem.CreateMenu(menu,m_defmenuid);
	POINT p;p.x = (LONG)pt.x;p.y = (LONG)pt.y;
	::ClientToScreen(m_hwnd,&p);
	UINT flag = TPM_RETURNCMD | TPM_NONOTIFY | TPM_LEFTBUTTON;
	switch(pos){
	case pos_leftbottom:
		flag |= (TPM_LEFTALIGN | TPM_BOTTOMALIGN);
		break;
	case pos_lefttop:
		flag |= (TPM_LEFTALIGN | TPM_TOPALIGN);
		break;
	case pos_rightbottom:
		flag |= (TPM_RIGHTALIGN | TPM_BOTTOMALIGN);
		break;
	case pos_righttop:
		flag |= (TPM_RIGHTALIGN | TPM_TOPALIGN);
		break;
	}
	int rv = ::TrackPopupMenu(menu,flag,p.x,p.y,0,m_hwnd,NULL);
	if(!rv)SetLastError();
	Destroy(menu);
	return rv;
}

Window::Window(HWND hwnd)
: Wnd(hwnd),m_menu(hwnd),
m_caret(hwnd),m_ime(hwnd),
m_hscroll(hwnd),m_vscroll(hwnd),
m_hdc_from_graphic(NULL)
{
	if(hwnd){
		HDC hdc = ::GetDC(hwnd);
		::GetTextMetrics(hdc,&m_textmetric);
		::ReleaseDC(hwnd,hdc);
		m_fontdata = FontData(m_fontdata.fontname,m_textmetric.tmHeight,m_textmetric.tmWeight,m_textmetric.tmAveCharWidth);
	}
}

Window::Window(Wnd *wnd)
: Wnd(wnd),m_menu(wnd),
m_caret(wnd),m_ime(wnd),
m_hscroll(wnd),m_vscroll(wnd),
m_hdc_from_graphic(NULL)
{
	if(GetHWND()){
		HDC hdc = ::GetDC(GetHWND());
		::GetTextMetrics(hdc,&m_textmetric);
		::ReleaseDC(GetHWND(),hdc);
		m_fontdata = FontData(m_fontdata.fontname,m_textmetric.tmHeight,m_textmetric.tmWeight,m_textmetric.tmAveCharWidth);
	}
}

void Window::Invalidate(Rect r){
	RECT rc;
	rc.top = (LONG)r.top;rc.bottom = (LONG)r.bottom;
	rc.left = (LONG)r.left;rc.right = (LONG)r.right;
	::InvalidateRect(m_hwnd,&rc,FALSE);
}

void Window::ScrollWindow(intptr_t x, intptr_t y){
	::ScrollWindowEx(m_hwnd,(int)x,(int)y,NULL,NULL,NULL,NULL,SW_INVALIDATE);
}

void Window::SetFont(FontData f){
//	fontsize = 2 * (fontsize / 2);
	m_fontdata = f;
	HFONT font = ::CreateFont((int)m_fontdata.height, (int)m_fontdata.avewidth,0,0, (int)m_fontdata.weight,FALSE,FALSE,FALSE,DEFAULT_CHARSET,OUT_DEFAULT_PRECIS,CLIP_DEFAULT_PRECIS,PROOF_QUALITY,FIXED_PITCH |FF_MODERN,m_fontdata.fontname.c_str());
	HDC hdc;
	if(m_hdc_from_graphic)hdc = m_hdc_from_graphic;
	else hdc = ::GetDC(m_hwnd);
	font = static_cast<HFONT>(::SelectObject(hdc,font));
	::GetTextMetrics(hdc,&m_textmetric);
	m_fontdata = FontData(m_fontdata.fontname,m_textmetric.tmHeight,m_textmetric.tmWeight,m_textmetric.tmAveCharWidth);
	::DeleteObject(::SelectObject(hdc,font));
	if(!m_hdc_from_graphic)::ReleaseDC(m_hwnd,hdc);
}

Size Window::GetTextExtent(const TCHAR *s,size_t len){
	HDC hdc;
	if(m_hdc_from_graphic)hdc = m_hdc_from_graphic;
	else HDC hdc = ::GetDC(m_hwnd);
	SIZE size;
	if(!::GetTextExtentPoint32(hdc,s,(int)len,&size))size.cx = -1;
	if(!m_hdc_from_graphic)::ReleaseDC(m_hwnd,hdc);
	return Size(size.cx,size.cy);
}

ablib::string Wnd::GetTitle(){
	int len = ::GetWindowTextLength(m_hwnd);
	TCHAR *buf = new TCHAR[len + 1];
	::GetWindowText(m_hwnd,buf,len + 1);
	ablib::string rv(buf);
	delete[] buf;
	return rv;
}

bool Wnd::EnumChildWindows(std::vector<Wnd> &rv){
	return ::EnumChildWindows(m_hwnd,sEnumWindowsProc,reinterpret_cast<LPARAM>(&rv)) != FALSE;
}

BOOL Wnd::sEnumWindowsProc(HWND hwnd,LPARAM lParam){
	std::vector<Wnd> *wnd = reinterpret_cast<std::vector<Wnd>*>(lParam);
	wnd->push_back(hwnd);
	return TRUE;
}

Window::msgresult Wnd::MsgBox(const TCHAR *msg,const TCHAR *title,msgstyle style){
	int type;
	switch(style){
	case mb_ok:				type = MB_OK;			break;
	case mb_okcancel:		type = MB_OKCANCEL;		break;
	case mb_retrycancel:	type = MB_RETRYCANCEL;	break;
	case mb_yesno:			type = MB_YESNO;		break;
	case mb_yesnocancel:	type = MB_YESNOCANCEL;	break;
	default:return id_failed;
	}
	switch(::MessageBox(m_hwnd,msg,title,type)){
	case IDOK:		return id_ok;
	case IDCANCEL:	return id_cancel;
	case IDYES:		return id_yes;
	case IDNO:		return id_no;
	default:		return id_failed;
	}
}

bool CommonDlg::OpenFileDlg(ablib::string &rv,const TCHAR *filter,const TCHAR *title,const TCHAR *def){
	OPENFILENAME ofn;
	TCHAR file[4096];
	
	::lstrcpy(file,_T(""));
	::ZeroMemory(&ofn,sizeof(OPENFILENAME));
	ofn.lStructSize = sizeof(OPENFILENAME);
	ofn.hwndOwner = m_hwnd;
	ofn.lpstrFilter = filter;
	ofn.lpstrTitle = title;
	ofn.lpstrFile = file;
	ofn.lpstrInitialDir = def;
	ofn.nMaxFile = 4095;
	ofn.Flags = OFN_FILEMUSTEXIST | OFN_HIDEREADONLY;

	if(::GetOpenFileName(&ofn) == FALSE)return false;
	rv = file;
	return true;
}

bool CommonDlg::SaveFileDlg(ablib::string &rv,const TCHAR *filter,const TCHAR *title,const TCHAR *def){
	OPENFILENAME ofn;
	TCHAR file[4096];
	
	::lstrcpy(file,_T(""));
	::ZeroMemory(&ofn,sizeof(OPENFILENAME));
	ofn.lStructSize = sizeof(OPENFILENAME);
	ofn.hwndOwner = m_hwnd;
	ofn.lpstrFilter = filter;
	ofn.lpstrTitle = title;
	ofn.lpstrFile = file;
	ofn.lpstrInitialDir = def;
	ofn.nMaxFile = 4095;
	ofn.Flags = OFN_HIDEREADONLY;

	if(::GetSaveFileName(&ofn) == FALSE)return false;
	rv = file;
	return true;
}

bool CommonDlg::SelectDir(ablib::string &r,const TCHAR *title,const TCHAR *def){
	BROWSEINFO bi;
	::ZeroMemory(&bi,sizeof(BROWSEINFO));

	TCHAR rv[4096];
	::lstrcpy(rv,_T(""));

	bi.hwndOwner = m_hwnd;
	bi.pszDisplayName = rv;
	bi.lpszTitle = title;
	bi.ulFlags = BIF_RETURNONLYFSDIRS;
	bi.lpfn = SelectDirProc;
	bi.lParam = reinterpret_cast<LPARAM>(def);

	LPITEMIDLIST retlist;
	retlist = SHBrowseForFolder(&bi);
	if(retlist){
		if(SHGetPathFromIDList(retlist,rv)){
			CoTaskMemFree(retlist);
			r = rv;
			return true;
		}else return false;
	}else return false;
}

int CommonDlg::SelectDirProc(HWND hwnd,UINT uMsg,LPARAM lParam,LPARAM data){
	switch(uMsg){
	case BFFM_INITIALIZED:
		if(data)::SendMessage(hwnd,BFFM_SETSELECTION,TRUE,data);
		return 0;
	default:
		return 0;
	}
}

}//namespace window
}//namespaca ablib
